/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.  

  This program is free software; you can redistribute it and/or	
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  $Header: /home/cvsroot/stage4/include/client_pamauth.h,v 1.3 2001/06/03 06:06:45 koni Exp $ 
  $Author: koni $ $Date: 2001/06/03 06:06:45 $ $Name:  $
  $Revision: 1.3 $

*/
   
#ifndef _CLIENT_PAMAUTH_H
#define _CLIENT_PAMAUTH_H
#include "slan_session.h"

/* It is unclear to even myself why the hell there are two of these, and why
they should be different. Perhaps if I hadn't created such an assinine
framework, there wouldn't be this weirdness. */
#define CLIENT_AUTH_TIMEMAX (60000000)
#define CLIENTAUTH_TIMEOUT (10000000)

/* TLV type encodings 
   There are possibly several PAM_MSG, MSG_DATA, and PAM_RESPONSE,
   type-length-value encodings in a single message, and we need to know the
   differences, so each one is PAM_MSG + message index, and likewise for the
   others.  */
enum { NUM_MSG = 0,
       PAM_MSG = 0x1000, MSG_DATA = 0x2000, PAM_RESPONSE = 0x4000 };

/* Client authentication message subtypes */
enum { AUTH_REQUEST = 0, AUTH_READY, AUTH_SKIP, PAM_SENDCLIENT, 
       PAM_CLIENTRESP, AUTH_OKAY, AUTH_REJECT };

/* PAM message tags, any message with tag of SLANPAM_UNKNOWN is ignored. */
enum { PROMPT_ECHO = 0, PROMPT_NOECHO, INFORMATION, WARNING, SLANPAM_UNKNOWN, N_PAMMSG_TYPES };

/* The "dialog" function will get an array of structures, which contain a
message or prompt to display, and a tag (type) which indicates whether or not
user input is expected, whether or not to echo the user's response, etc. These
tags are defined above. */
typedef struct {
  int type;
  unsigned char *data;
} client_authmsg_t;

/* The client's response to the pam dialog is returned as an array of 
   pointers to responses, some of which may be NULL if the corresponding 
   PAM message was not a prompting message. */
typedef unsigned char * client_resp_t;

int server_clientauth_initiate(slan_session_t *session_obj);
int client_clientauth_wait(slan_session_t *session_obj);

#endif

