/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
  $Header: /home/cvsroot/stage4/platform/windows/driver/slanmac/slan.h,v 1.1 2001/06/05 16:02:13 lucasm Exp $
  $Author: lucasm $ $Date: 2001/06/05 16:02:13 $ $Name:  $
  $Revision: 1.1 $
 
*/

#include <ndis.h>
#include <devioctl.h>

#define WANTVXDWRAPS

#include <basedef.h>
#include <vmm.h>
#include <debug.h>
#include <vxdwraps.h>
#include <vwin32.h>
#include <winerror.h>

/*
 * use windows standards on how to declare structures,
 * this makes it easier (?) to interoperate with their functions.
 */

typedef struct tagSLAN_BINDING
{
	NDIS_HANDLE NdisBindingContextFromProtocol; // given to us at openadapter()
	struct tagSLAN_ADAPTER *Adapter;			// point back to our adapter
	struct tagSLAN_BINDING *Next;			    // next binding...
} SLAN_BINDING, *PSLAN_BINDING;

struct waiting_buffer {
   OVERLAPPED *lpoOverlapped;  // overlapped IO garbage
   UCHAR *buffer;	       // pointer to the user-passed buffer
   DWORD buffer_length;        // length of this buffer
   DWORD *lpcbBytesReturned;   // pointer to bytes returned
};

typedef struct tagSLAN_ADAPTER
{
	PSLAN_BINDING Bindings;		   // protocol bindings to this adapter
	NDIS_HANDLE NdisAdapterHandle; // handle ndis gives us at AddAdapter()
	BOOL HasLink;                  // has a user-mode link set up
	PVOID ScratchBuffer;           // scratch buffer for constructing packet
	UINT BufferSize;               // size of my internal circular buffer
	PUCHAR Buffer;                 // pointer to internal circular buffer;
	UINT BufferHead;               // buffer + bufferhead = head of buffer
	UINT BufferTail;               // (should be self explanatory)
	UINT BufferLastByte;           // Last byte of data in buffer
	BOOL HasWaitingBuffer;		   // is a buffer waiting?
	struct waiting_buffer WaitingBuffer; // if so, this is it..
	BOOL HaveDataEvent;			   // do I have an event to signal when there is data?
	DWORD DataEvent;			   // VXD data event!
} SLAN_ADAPTER, *PSLAN_ADAPTER;

typedef struct tagSLAN_DRIVER 
{
   PSLAN_ADAPTER TheAdapter;		  // pointer to our one adapter context
   NDIS_HANDLE NdisWrapperHandle; // handle ndis gives us at InitializeWrapper()
   NDIS_HANDLE NdisMacHandle;     // handle ndis gives us at RegisterMac()
} SLAN_DRIVER, *PSLAN_DRIVER;

struct slan_header { // structure stuck before each packet
	DWORD framelength;	// length of the frame 
	DWORD length;		// length of the actual data following (not including length of this packet)
};

#define SLAN_OVERHEAD		   160	// overhead caused by slan-ness (FIXME!)
#define ETHERNET_HEADER_LEN	   14	// self explanatory
#define SCRATCH_BUFFER_LENGTH  2000 // this is hideously gross, but I need some form of 
                                    // internal scratch buffer to make a flat packet
                                    // from an NDIS_PACKET grossness.
#define DEFAULT_BUFFER_SIZE 65536   // default kernel->user buffer size
									// currently this isn't changeable.
									// maybe allow shrink/grow on future release?

// (from devioctl.h)
// Define the various device type values.  Note that values used by Microsoft
// Corporation are in the range 0-32767, and 32768-65535 are reserved for use
// by customers.

// I'm not sure this really matters at all under 9x, unless we were also registering
// to provide USB device services or something.
// Still, why not.

#define FILE_DEVICE_SLAN    0xA5B4 //42420 -- magic number?
#define IOCTL_SLAN_OPEN		CTL_CODE(FILE_DEVICE_SLAN, METHOD_BUFFERED, 0x01, FILE_ANY_ACCESS)
#define IOCTL_SLAN_CLOSE	CTL_CODE(FILE_DEVICE_SLAN, METHOD_BUFFERED, 0x02, FILE_ANY_ACCESS)
#define IOCTL_SLAN_READ		CTL_CODE(FILE_DEVICE_SLAN, METHOD_BUFFERED, 0x03, FILE_ANY_ACCESS)
#define IOCTL_SLAN_WRITE	CTL_CODE(FILE_DEVICE_SLAN, METHOD_BUFFERED, 0x04, FILE_ANY_ACCESS)

////////// DEBUG STUFF BELOW! 

#if (defined(DBG))

#define DBG_FUNC(F)		static const char __FUNC__[] = F;


#define DBG_ENTER()		{ DbgPrint("SLANMAC: >>>>> %s \n", __FUNC__); }

#define DBG_LEAVE(R)	{ DbgPrint("SLANMAC: <<<<< %s returning (0x%08X) \n", __FUNC__, (R)); }

#define DBG_BREAK()		{ DbgPrint ("SLANMAC: BreakPoint in %s at line %d \n", __FUNC__, __LINE__); \
							  DbgBreakPoint(); \
						}

#define DBG_PRINT(S)	{ DbgPrint("SLANMAC: %s: ", __FUNC__); DbgPrint S }

#define DBG_DISPLAY(S)	{	DbgPrint("DBG TRACE: "); DbgPrint S }

#ifdef ASSERT
#undef ASSERT
#endif

#define ASSERT(C)     if (!(C)) { \
                          DbgPrint("!---%s: ASSERT(%s) FAILED!\n%s #%d\n", \
                                         __FUNC__, #C, __FILE__, __LINE__); \
                          DbgBreakPoint(); \
                      }

#else	// ( defined(DBG) && (DBG != 0) )

#define DBG_FUNC(F)
#define DBG_ENTER()
#define DBG_LEAVE(R)
#define DBG_BREAK()
#define DBG_PRINT(S)
#define DBG_LOCAL(V, P)
#endif

//// our memory can be anywhere
static const NDIS_PHYSICAL_ADDRESS HighestAcceptableMax = NDIS_PHYSICAL_ADDRESS_CONST(-1,-1);

/// function prototypes;

NDIS_STATUS NDIS_API SLAN_AddAdapter(
    NDIS_HANDLE MacMacContext,
    NDIS_HANDLE ConfigurationHandle,
    PNDIS_STRING AdapterName);

NDIS_STATUS NDIS_API 
SLAN_OpenAdapter(
    OUT PNDIS_STATUS OpenErrorStatus,  // Token ring. Set to success.
    OUT NDIS_HANDLE *MacBindingHandle, // This is what I return (my internal structure?)
    OUT PUINT SelectedMediumIndex,	   // media type
    IN PNDIS_MEDIUM MediumArray,       // array of media types
    IN UINT MediumArraySize,		   // size of this array
    IN NDIS_HANDLE NdisBindingContext, // binding context for this protocol (?)
    IN NDIS_HANDLE MacAdapterContext,  // SLAN_ADAPTER 
    IN UINT OpenOptions,               // ??? 
    IN PSTRING AddressingInformation OPTIONAL // hareware specific info
    );

NDIS_STATUS NDIS_API 
SLAN_CloseAdapter(
    IN NDIS_HANDLE MacBindingHandle
    );

NDIS_STATUS NDIS_API
SLAN_Request(
    IN NDIS_HANDLE MacBindingHandle,
    IN PNDIS_REQUEST NdisRequest
    );

VOID NDIS_API
SLAN_RemoveAdapter(
    IN PVOID MacAdapterContext
    );

VOID NDIS_API
SLAN_Unload(
    IN NDIS_HANDLE MacMacContext
    );

NDIS_STATUS NDIS_API
SLAN_Reset(
    IN NDIS_HANDLE MacBindingHandle
    );

NDIS_STATUS NDIS_API
SLAN_Send(
    IN NDIS_HANDLE Context,
    IN PNDIS_PACKET Packet
    );

NDIS_STATUS NDIS_API
SLAN_TransferData(
    IN NDIS_HANDLE MacBindingHandle,
    IN NDIS_HANDLE MacReceiveContext,
    IN UINT ByteOffset,              
    IN UINT BytesToTransfer,         
    OUT PNDIS_PACKET Packet,         
    OUT PUINT BytesTransferred       
    );

NDIS_STATUS NDIS_API
SLAN_QueryGlobalStatistics(
    IN NDIS_HANDLE MacContext,
    IN PNDIS_REQUEST NdisRequest
    );

SLAN_DRIVER SlanDriver;

NDIS_STATUS deal_with_sent_buffer(PSLAN_ADAPTER adapter, PVOID data, UINT datalen);
DWORD LockMemory(DWORD addr, DWORD length);
DWORD UnlockMemory(DWORD address, DWORD length);
DWORD SetVXDEvent(DWORD event);
