/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
  $Header: /home/cvsroot/stage4/platform/windows/driver/slanmac/slanioctl.c,v 1.1 2001/06/05 16:02:13 lucasm Exp $
  $Author: lucasm $ $Date: 2001/06/05 16:02:13 $ $Name:  $
  $Revision: 1.1 $
 
*/

#include "slan.h"

#pragma VxD_LOCKED_CODE_SEG
#pragma VxD_LOCKED_DATA_SEG

DWORD User_Read(PSLAN_ADAPTER adapter, 
				PDIOCPARAMETERS pdioc);
DWORD User_Write(PSLAN_ADAPTER adapter, 
				PDIOCPARAMETERS pdioc);

int CreateOpenInstance(PSLAN_ADAPTER adapter, PDIOCPARAMETERS pdioc);
int DestroyOpenInstance(adapter);

// RETURN VALUES:
// -1: This is an asynchronous operation. 
//     (does not return control until VWIN32_DIOCompletionRoutine)
// 0: success
// anything else: error
DWORD _stdcall SLAN_W32_DeviceIOControl(
    DWORD hDevice, //ebx, handle to the VMM (do we care?)
    PDIOCPARAMETERS pdioc // esi (gross evil structure from satan)
	)
{
	DBG_FUNC("DeviceIOControl")
    DWORD dwRetVal = 0;
	PSLAN_ADAPTER adapter = SlanDriver.TheAdapter;

	DBG_ENTER();

	// we must respond to a DIOC_OPEN with zero to tell VMM we support this!
	if(pdioc->dwIoControlCode == DIOC_OPEN)
	{
		DBG_PRINT(("DIOC_OPEN\n"););
		DBG_LEAVE(0);
		return 0;
	}

	// what exactly does "closehandle" mean?
	// I'm assuming it means someone closed our CreateFile() instance.
	// That means a) the OS is cleaning up because there was a crash
	// or b) it's us closing. But, we try to clean up with our own IOCTLs anyway.
	if(pdioc->dwIoControlCode == DIOC_CLOSEHANDLE)
	{
		DBG_PRINT(("DIOC_CLOSEHANDLE\n"););
		if(adapter && adapter->HasLink)
		{
			DBG_PRINT(("has_link!!\n"););
			DestroyOpenInstance(adapter);
		}
		DBG_LEAVE(0);
		return 0;
	}

    if(adapter == NULL)
		return 1; // error!

	switch(pdioc->dwIoControlCode)
	{
		case IOCTL_SLAN_CLOSE:
			DestroyOpenInstance(adapter);
			return 0;

		case IOCTL_SLAN_OPEN:
			if(CreateOpenInstance(adapter, pdioc))
				dwRetVal = 1;
			break;

		case IOCTL_SLAN_READ:
			dwRetVal = User_Read(adapter, pdioc);
			break;

		case IOCTL_SLAN_WRITE:
			dwRetVal = User_Write(adapter, pdioc);
			break;

		default:
			// what is this? :)
			dwRetVal = 1;
			break;
	}

	DBG_LEAVE(dwRetVal);
	return dwRetVal;
}

int DestroyOpenInstance(PSLAN_ADAPTER adapter)
{
	DBG_FUNC("DestroyOpenInstance")

	DBG_ENTER();

	if(adapter->HasLink == FALSE)
		return 0;

	if(adapter->HasWaitingBuffer)
	{
		UnlockMemory((DWORD) adapter->WaitingBuffer.buffer, adapter->WaitingBuffer.buffer_length);
		UnlockMemory((DWORD) adapter->WaitingBuffer.lpoOverlapped, sizeof(OVERLAPPED));
		UnlockMemory((DWORD) adapter->WaitingBuffer.lpcbBytesReturned, sizeof(DWORD));
		adapter->HasWaitingBuffer = FALSE;
	}

	if(adapter->HaveDataEvent == TRUE)
	{
		DWORD devent = adapter->DataEvent;

		_asm mov eax, devent
		VxDCall(_VWIN32_CloseVxDHandle);

		adapter->HaveDataEvent = FALSE;
	}

	adapter->HasLink = FALSE;
	NdisFreeMemory(adapter->ScratchBuffer, SCRATCH_BUFFER_LENGTH, 0);
    NdisFreeMemory(adapter->Buffer, DEFAULT_BUFFER_SIZE, 0);

	DBG_LEAVE(0);
	return 0;
}

int CreateOpenInstance(PSLAN_ADAPTER adapter, PDIOCPARAMETERS pdioc)
{
	DBG_FUNC("CreateOpenInstance")
	NDIS_STATUS status;

	DBG_ENTER();

	ASSERT(adapter->HasLink == FALSE);

	if(adapter->HasLink != FALSE)
		return -1;

	if(!(pdioc->cbInBuffer == 0 || pdioc->cbInBuffer == sizeof(DWORD)))
		return -1;

	status = NdisAllocateMemory(&adapter->ScratchBuffer, SCRATCH_BUFFER_LENGTH, 0,
		HighestAcceptableMax);

	if(status != NDIS_STATUS_SUCCESS)
	{
		DBG_PRINT(("Couldn't allocate scratch_buffer\n"););
		DBG_LEAVE(-1);
		return -1;
	}

	status = NdisAllocateMemory(&adapter->Buffer, DEFAULT_BUFFER_SIZE, 0,
		HighestAcceptableMax);

	if(status != NDIS_STATUS_SUCCESS)
	{
		NdisFreeMemory(adapter->ScratchBuffer, SCRATCH_BUFFER_LENGTH, 0);
		DBG_PRINT(("Couldn't allocate buffer\n"););
		DBG_LEAVE(-1);
		return -1;
	}

	if(pdioc->cbInBuffer == sizeof(DWORD))
	{
		DWORD *evnt = (DWORD *) pdioc->lpvInBuffer;
		adapter->HaveDataEvent = TRUE;
		adapter->DataEvent = *evnt;
	}

	adapter->BufferSize = DEFAULT_BUFFER_SIZE;
	adapter->BufferHead = 0;
	adapter->BufferTail = 0;
	adapter->BufferLastByte = 0;

	adapter->HasLink = TRUE;
	DBG_LEAVE(0);
	return 0;
}

void SLANMoveMemory(PVOID dest, PVOID src, UINT len)
{
	DBG_FUNC("SLANMoveMemory")
	UINT nblocks, i;

	ASSERT(!(len & 3)); // we 4-byte aligned for this, did we not?
	ASSERT(!((DWORD)src & 3)); // can't do this with that!

	nblocks = len / 4;

	for(i = 0; i < nblocks; i++)
	{
		*((PULONG)dest)++ = *((PULONG)src)++;
	}
}

void SLANMoveMemoryNA(PVOID dest, PVOID src, UINT len)
{
	DBG_FUNC("SLANMoveMemory")
	UINT nblocks, i;

	ASSERT(!((DWORD)src & 3)); // can't do this with that!

	nblocks = len / 4;

	for(i = 0; i < nblocks; i++)
	{
		*((PULONG)dest)++ = *((PULONG)src)++;
	}

	nblocks = len & 0x03;

	for(i = 0; i < nblocks; i++)
	{
		*((PUCHAR)dest)++ = *((PUCHAR)src)++;
	}
}


void VXDINLINE DoneWithIO(DWORD parm)
{
	_asm mov ebx, [parm]
    VxDCall(VWIN32_DIOCCompletionRoutine)
}

NDIS_STATUS deal_with_sent_buffer(PSLAN_ADAPTER adapter, PVOID data, UINT datalen)
{
	DBG_FUNC("deal_with_sent_buffer")
	UINT head, tail, lbyte, spacereq;
	struct slan_header *slheader;
	PUCHAR curbuf;
	UINT copied_len;

	DBG_ENTER();

	spacereq = datalen + sizeof(struct slan_header);

	if(adapter->HasWaitingBuffer == FALSE)
	{
		// nothing is waiting for this data already, we need to buffer it!
        ASSERT(adapter->BufferSize);

		// this implements a "circular" buffer.
		// a packet cannot wrap around the buffer.

		DBG_PRINT(("No waiting buffer...\n"););

		head = adapter->BufferHead;
		tail = adapter->BufferTail;
		lbyte = adapter->BufferLastByte;

		if(tail + spacereq >= adapter->BufferSize) // can't stick it at the end
		{
			if(head <= spacereq) // can't stick it at the beginning
			{
				DBG_PRINT(("Buffer is full, dropping packet.\n"););
				return NDIS_STATUS_NOT_ACCEPTED;
			}
			else
			{
				/* set the tail to the beginning of the buffer */
				tail = 0;
			} 
		}

		if( (tail < head) && (tail + spacereq >= head) )
		{
			// buffer is full.
			DBG_PRINT(("Buffer is full, dropping packet.\n"););
			return NDIS_STATUS_NOT_ACCEPTED;
		}

		curbuf = adapter->Buffer + tail;
		slheader = (struct slan_header *) curbuf;

		slheader->framelength = datalen;
		slheader->length = datalen;

		NdisMoveMemory((PVOID)(curbuf + sizeof(struct slan_header)),
			data, datalen);

		tail += spacereq;
		/* round up to nearest word, so we don't copy off boundaries */
		/* is this correct? should I round up to the nearest dword? */
		while(tail & 3)
		{
			slheader->framelength++;
			tail++;
		}
		if(tail > head)
			lbyte = tail;
		adapter->BufferTail = tail;
		adapter->BufferLastByte = lbyte;
		// the head does not change, duh :)
		if(adapter->HaveDataEvent)
		{
			DWORD devent = adapter->DataEvent;
			extern void __cdecl SetUserEvent(void);

			_asm mov edx, devent
			SetUserEvent();
		}
		return NDIS_STATUS_SUCCESS;
	}


	//DBG_PRINT(("Waiting buffer...!\n"););
	// ooo, a waiting buffer...
	copied_len = 0;

	if(adapter->WaitingBuffer.buffer_length >= spacereq)
	{
		//DBG_PRINT(("Copying buffer...\n"););

		curbuf = adapter->WaitingBuffer.buffer;
		slheader = (struct slan_header *) curbuf;

		slheader->framelength = datalen;
		slheader->length = datalen;

		SLANMoveMemoryNA(curbuf + sizeof(struct slan_header), data, datalen);

		copied_len = spacereq;
	}

	//DBG_PRINT(("Setting stuff.. [%d]\n", copied_len););

	*adapter->WaitingBuffer.lpcbBytesReturned = copied_len;
	adapter->WaitingBuffer.lpoOverlapped->O_InternalHigh = copied_len;

	//DBG_PRINT(("Calling VWIN32\n"););

	DoneWithIO(adapter->WaitingBuffer.lpoOverlapped->O_Internal);

	//DBG_PRINT(("Unlocking Memory...\n"););

	UnlockMemory((DWORD) adapter->WaitingBuffer.buffer, adapter->WaitingBuffer.buffer_length);
	UnlockMemory((DWORD) adapter->WaitingBuffer.lpoOverlapped, sizeof(OVERLAPPED));
	UnlockMemory((DWORD) adapter->WaitingBuffer.lpcbBytesReturned, sizeof(DWORD));
	adapter->HasWaitingBuffer = FALSE;

	//DBG_PRINT(("DONE!\n"););

	return NDIS_STATUS_SUCCESS;
}

// called when we have a read command from usermode!
DWORD User_Read(PSLAN_ADAPTER adapter, 
				PDIOCPARAMETERS pdioc)
{
	DBG_FUNC("User_Read")
	UINT head; // head of the buffer
	UINT tail; // tail of the buffer
	UINT lbyte; // last byte in the buffer (head can be > tail!) 
//	UINT copytmp, copylen, totalcopylen;
	UINT thispktlen; // not needed, but used for simple clarity
	PUCHAR userbuffer;
	UINT userbufferlen;
	DBG_ENTER();

	ASSERT(adapter->HasWaitingBuffer == FALSE);

	if(adapter->HasWaitingBuffer)
	{
		// strange. error.
		return 1;
	}

	if(pdioc->cbOutBuffer < adapter->BufferSize) // dangerously small user buffer?
	{
		DBG_PRINT(("buffer too small!\n"););
		DBG_LEAVE(1);
		return 1;
	}

	if(adapter->BufferHead == adapter->BufferTail)
	{
		// there is no data waiting to be read.

		if(pdioc->lpoOverlapped == NULL)
		{
			*(DWORD *)(pdioc->lpcbBytesReturned) = 0;
			return 0; // cannot do overlapped IO
		}

		adapter->HasWaitingBuffer = TRUE;
		adapter->WaitingBuffer.buffer = (PVOID) LockMemory(pdioc->lpvOutBuffer, pdioc->cbOutBuffer);
		if(adapter->WaitingBuffer.buffer == NULL)
		{
			DBG_PRINT(("Memory locking FAILED! (1)\n"););
			return 1;
		}

		adapter->WaitingBuffer.lpoOverlapped = (PVOID) LockMemory(pdioc->lpoOverlapped, sizeof(OVERLAPPED));
		if(adapter->WaitingBuffer.lpoOverlapped == NULL)
		{
			UnlockMemory((DWORD) adapter->WaitingBuffer.buffer, pdioc->cbOutBuffer);
			DBG_PRINT(("Memory locking FAILED! (2)\n"););
			return 1;
		}

  	    adapter->WaitingBuffer.lpcbBytesReturned = (PVOID) LockMemory(pdioc->lpcbBytesReturned, sizeof(DWORD));
		if(adapter->WaitingBuffer.buffer == NULL)
		{
			UnlockMemory((DWORD) adapter->WaitingBuffer.buffer, pdioc->cbOutBuffer);
			UnlockMemory((DWORD) adapter->WaitingBuffer.lpoOverlapped, sizeof(OVERLAPPED));
			DBG_PRINT(("Memory locking FAILED! (3)\n"););
			return 1;
		}

		adapter->WaitingBuffer.buffer_length = pdioc->cbOutBuffer;
		return -1; // start overlappedness, ooo
	}

	// we do not need to lock the memory here as we
	// are not running as an ndis callback.
	// we can wait for this to be swapped in.
	// we however cannot use NdisMoveMemory to move memory, it
	// seems to have issues.

	userbufferlen = pdioc->cbOutBuffer;
	userbuffer = (PUCHAR) pdioc->lpvOutBuffer;

	head = adapter->BufferHead;
	tail = adapter->BufferTail;
	lbyte = adapter->BufferLastByte;

	if(tail > head) // normal linear buffer?
	{
		thispktlen = (tail - head); 
		if(thispktlen < userbufferlen) // enough space to copy it all?
		{
			SLANMoveMemory(userbuffer, adapter->Buffer + head, thispktlen);
			adapter->BufferHead += thispktlen;
			*(DWORD *)(pdioc->lpcbBytesReturned) = thispktlen;
			DBG_LEAVE(0);
			return 0;
		}
	}
	else // head > tail, wacky buffer
	{
		thispktlen = (lbyte - head);
		if(thispktlen < userbufferlen)  // in this segment from head -> last?
		{
			SLANMoveMemory(userbuffer, adapter->Buffer + head, thispktlen);
			adapter->BufferLastByte = tail;
			adapter->BufferHead = 0;
			*(DWORD *)(pdioc->lpcbBytesReturned) = thispktlen;
  			DBG_LEAVE(0);
			return 0;
		}
	}

	// if after this read, there is still data in the buffer...
	// set the event again!

	if(adapter->HaveDataEvent && (adapter->BufferHead != adapter->BufferTail))
	{
		DWORD devent = adapter->DataEvent;
		extern void __cdecl SetUserEvent(void);

		_asm mov edx, devent
		SetUserEvent();
	}

	// techically, we should never get here.
	// this code is to deal with a user buffer of less than the amount of our internal buffer.
	// however, it is cast in stone up above that the user buffer MUST be at least the size
	// of our internal buffer.

	DBG_LEAVE(1);
	return 1;

	/* (see comment above)
	copytmp = head;
	totalcopylen = 0;
	while(1)
	{
		if(head == tail) // out of data?
			break;

		if(head == lbyte) // has the head moved up to be the last byte in the buffer?
		{
			// yes, copytmp is the actual head, head is the tail of that segment.
			// that means head should become zero when done.
			// copytemp < head, guaranteed.
			ASSERT(copytmp < head);

			SLANMoveMemory(userbuffer, adapter->Buffer + copytmp, head - copytmp);
			userbuffer += (head - copytmp);
			head = 0;
			adapter->BufferHead = 0;
			copytmp = 0;
		}

		copylen = ((struct slan_header *) (adapter->Buffer + head))->framelength +
			sizeof(struct slan_header);

		ASSERT(!(copylen & 3)) // framelen should be word aligned
		//while(copylen & 3) // 4-byte align...
		//	copylen++;

		if(totalcopylen + copylen > userbufferlen)
			break; // out of user buffer space?

		totalcopylen += copylen;
		head += copylen;
	}

	SLANMoveMemory(userbuffer, adapter->Buffer + copytmp, head - copytmp);
	adapter->BufferHead += head - copytmp;

	*(DWORD *)(pdioc->lpcbBytesReturned) = totalcopylen;

	DBG_LEAVE(0);
	return 0;
	*/
}

// called when we have a read command from usermode!
DWORD User_Write(PSLAN_ADAPTER adapter, 
				PDIOCPARAMETERS pdioc)
{
	DBG_FUNC("User_Write")
	PSLAN_BINDING sopen;
	NDIS_STATUS status;
	UCHAR *outbuffer;
	UINT outbufferlength;

	DBG_ENTER()

	outbuffer = (PVOID) LockMemory(pdioc->lpvInBuffer, pdioc->cbInBuffer);
	if(outbuffer == NULL)
	{
		DBG_PRINT(("Memory locking FAILED! (W)\n"););
		return 1;
	}

	outbufferlength = pdioc->cbInBuffer;
	*(DWORD *)pdioc->lpcbBytesReturned = 0;

	for(sopen = adapter->Bindings; sopen; sopen = sopen->Next)
	{
		NdisIndicateReceive(&status, 
			sopen->NdisBindingContextFromProtocol,
			(NDIS_HANDLE)adapter,
			outbuffer,
			ETHERNET_HEADER_LEN,
			outbuffer + ETHERNET_HEADER_LEN,
			outbufferlength - ETHERNET_HEADER_LEN,
			outbufferlength - ETHERNET_HEADER_LEN);

		if(status != NDIS_STATUS_NOT_ACCEPTED)
		{
			NdisIndicateReceiveComplete(sopen->NdisBindingContextFromProtocol);
		}
	}
	UnlockMemory((DWORD) outbuffer, outbufferlength);
	return 0;
}
