/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
  $Header: /home/cvsroot/stage4/platform/windows/driver/slanmac/slanmain.c,v 1.1 2001/06/05 16:02:13 lucasm Exp $
  $Author: lucasm $ $Date: 2001/06/05 16:02:13 $ $Name:  $
  $Revision: 1.1 $
 
*/

#include "slan.h"

//// global structure
SLAN_DRIVER SlanDriver = {0};
//// name of our driver
NDIS_STRING SlanDriverName = NDIS_STRING_CONST("SLANmac");

#pragma NDIS_INIT_FUNCTION(DriverEntry)

NTSTATUS NDIS_API DriverEntry(
    IN PDRIVER_OBJECT DriverObject,
    IN PUNICODE_STRING RegistryPath
    )
{
	DBG_FUNC("DriverEntry")
	NDIS_STATUS status;
	NDIS_MAC_CHARACTERISTICS mchar;

	DBG_ENTER();

	DBG_PRINT(("\n\nslanmac: Build date: "__DATE__" Time: "__TIME__"\n"););

	NdisZeroMemory(&SlanDriver, sizeof(SLAN_DRIVER));

	NdisInitializeWrapper(&SlanDriver.NdisWrapperHandle, DriverObject, RegistryPath, NULL);
	//Apparently this whole wrapper handle thing
	//is not used on 95/98, it's just a dummy function (maybe?)
	//It's always set to NULL.
	//ASSERT(SlanDriver.NdisWrapperHandle);

	NdisZeroMemory(&mchar, sizeof(NDIS_MAC_CHARACTERISTICS));

	mchar.MajorNdisVersion = NDIS_MAJOR_VERSION;
	mchar.MinorNdisVersion = NDIS_MINOR_VERSION;

	mchar.AddAdapterHandler = SLAN_AddAdapter;
	mchar.CloseAdapterHandler = SLAN_CloseAdapter;
	mchar.OpenAdapterHandler = SLAN_OpenAdapter;
	mchar.QueryGlobalStatisticsHandler = SLAN_QueryGlobalStatistics;
	mchar.RemoveAdapterHandler = SLAN_RemoveAdapter;
	mchar.ResetHandler = SLAN_Reset;
	mchar.RequestHandler = SLAN_Request;
	mchar.SendHandler = SLAN_Send;
	mchar.TransferDataHandler = SLAN_TransferData;
	mchar.UnloadMacHandler = SLAN_Unload;

	mchar.Name = SlanDriverName;

	NdisRegisterMac(&status, &SlanDriver.NdisMacHandle,
		SlanDriver.NdisWrapperHandle, (NDIS_HANDLE) &SlanDriver,
		&mchar, sizeof(NDIS_MAC_CHARACTERISTICS));

	if(status != NDIS_STATUS_SUCCESS)
	{
		DBG_PRINT(("!@!@#! NdisRegisterMac status=%xh\n", status););
        NdisTerminateWrapper(&SlanDriver.NdisWrapperHandle, NULL);
	}
	else
	{
		DBG_PRINT(("!!! Driver loaded successfully!\n"););
	}

	DBG_LEAVE(status);
	return status;
}

NDIS_STATUS NDIS_API SLAN_AddAdapter(
    NDIS_HANDLE MacMacContext,
    NDIS_HANDLE ConfigurationHandle,
    PNDIS_STRING AdapterName)
{
	DBG_FUNC("SLAN_AddAdapter")
	PSLAN_ADAPTER adapter;
    NDIS_ADAPTER_INFORMATION adapter_info;
	NDIS_STATUS status;

	DBG_ENTER();

	if(SlanDriver.TheAdapter != NULL)
	{
		DBG_PRINT(("AddAdapter called when adapter already exists. Dual adapter configuration??\n"););
		DBG_LEAVE(NDIS_STATUS_FAILURE);
		return NDIS_STATUS_FAILURE;
	}

	status = NdisAllocateMemory(&adapter, sizeof(SLAN_ADAPTER), 0, HighestAcceptableMax);
	if(status != NDIS_STATUS_SUCCESS)
	{
		DBG_PRINT(("!@!@#! Couldn't get memory for adapter, status=%xh\n", status););
		DBG_LEAVE(status);
		return status;
	}

	NdisZeroMemory(adapter, sizeof(SLAN_ADAPTER));
	adapter->HasLink = FALSE; // no link yet!
	adapter->HaveDataEvent = FALSE; // no data event!

	NdisZeroMemory(&adapter_info, sizeof(NDIS_ADAPTER_INFORMATION));
	adapter_info.AdapterType = NdisInterfaceInternal;

	status = NdisRegisterAdapter(&adapter->NdisAdapterHandle, 
		SlanDriver.NdisMacHandle, (NDIS_HANDLE) adapter,
		ConfigurationHandle, AdapterName, &adapter_info);

	if(status != NDIS_STATUS_SUCCESS)
	{
		DBG_PRINT(("!@!@#! NdisRegisterAdapter failed, status=%xh\n", status););
		NdisFreeMemory(adapter, sizeof(SLAN_ADAPTER), 0);
		DBG_LEAVE(status);
		return status;
	}

	SlanDriver.TheAdapter = adapter;

	return NDIS_STATUS_SUCCESS;
}

NDIS_STATUS NDIS_API 
SLAN_OpenAdapter(
    OUT PNDIS_STATUS OpenErrorStatus,  // Token ring. Set to success.
    OUT NDIS_HANDLE *MacBindingHandle, // This is what I return (my internal structure?)
    OUT PUINT SelectedMediumIndex,	   // media type
    IN PNDIS_MEDIUM MediumArray,       // array of media types
    IN UINT MediumArraySize,		   // size of this array
    IN NDIS_HANDLE NdisBindingContext, // binding context for this protocol (?)
    IN NDIS_HANDLE MacAdapterContext,  // SLAN_ADAPTER 
    IN UINT OpenOptions,               // ??? 
    IN PSTRING AddressingInformation OPTIONAL // hareware specific info
    )
{
	DBG_FUNC("SLAN_OpenAdapter")
	UINT i;
	PSLAN_BINDING pbinding;
	PSLAN_ADAPTER adapter = (PSLAN_ADAPTER) MacAdapterContext;
	NDIS_STATUS status;

	DBG_ENTER();

	// scan media array for ethernet, return if not found

	for(i = 0; i < MediumArraySize; i++)
	{
		if(MediumArray[i] == NdisMedium802_3)
		{
			*SelectedMediumIndex = i;
			break;
		}
	}

	if(i == MediumArraySize)
	{
		DBG_LEAVE(NDIS_STATUS_UNSUPPORTED_MEDIA);
		return NDIS_STATUS_UNSUPPORTED_MEDIA;
	}

	status = NdisAllocateMemory(&pbinding, sizeof(SLAN_BINDING), 0, HighestAcceptableMax);

	if(status != NDIS_STATUS_SUCCESS)
	{
		DBG_LEAVE(NDIS_STATUS_RESOURCES);
		return NDIS_STATUS_RESOURCES;
	}

	pbinding->Adapter = adapter;
	pbinding->NdisBindingContextFromProtocol = NdisBindingContext;
	pbinding->Next = adapter->Bindings;
	adapter->Bindings = pbinding;

	*MacBindingHandle = (NDIS_HANDLE) pbinding;

	DBG_LEAVE(NDIS_STATUS_SUCCESS);
	return NDIS_STATUS_SUCCESS;
}

NDIS_STATUS NDIS_API 
SLAN_CloseAdapter(
    IN NDIS_HANDLE MacBindingHandle
    )
{
	DBG_FUNC("SLAN_CloseAdapter")
	PSLAN_BINDING pbinding = (PSLAN_BINDING) MacBindingHandle, tmpb;
	PSLAN_ADAPTER adapter = pbinding->Adapter;

	/* remove it from the singly linked list of bindings */

	if(pbinding == adapter->Bindings)
	{
		adapter->Bindings = adapter->Bindings->Next;
	}
	else
	{
		tmpb = adapter->Bindings;

		while(tmpb->Next != pbinding)
		{
			tmpb = tmpb->Next;
		}

		tmpb->Next = pbinding->Next;
	}

	NdisFreeMemory(pbinding, sizeof(SLAN_BINDING), 0);

	DBG_LEAVE(NDIS_STATUS_SUCCESS);
	return NDIS_STATUS_SUCCESS;
}

VOID NDIS_API
SLAN_RemoveAdapter(
    IN PVOID MacAdapterContext
    )
{
	DBG_FUNC("SLAN_RemoveAdapter")
	PSLAN_ADAPTER adapter = (PSLAN_ADAPTER) MacAdapterContext;

	DBG_ENTER();

	ASSERT(adapter->Bindings == NULL);
	ASSERT(SlanDriver.TheAdapter == adapter);

	NdisFreeMemory(adapter, sizeof(SLAN_ADAPTER), 0);
	SlanDriver.TheAdapter = NULL;
	DBG_LEAVE(0);
}

VOID NDIS_API
SLAN_Unload(
    IN NDIS_HANDLE MacMacContext
    )
{
	DBG_FUNC("SLAN_Unload")
	NDIS_STATUS tmp;

	DBG_ENTER();

	NdisDeregisterMac(&tmp, SlanDriver.NdisMacHandle);
	SlanDriver.NdisMacHandle = NULL;

	NdisTerminateWrapper(SlanDriver.NdisWrapperHandle, NULL);
	SlanDriver.NdisWrapperHandle = NULL;

	DBG_LEAVE(0);
	return;
}

NDIS_STATUS NDIS_API
SLAN_Reset(
    IN NDIS_HANDLE MacBindingHandle
    )
{
	DBG_FUNC("SLAN_Reset")
	
	DBG_ENTER();
	DBG_LEAVE(NDIS_STATUS_SUCCESS);
	return NDIS_STATUS_SUCCESS;
}

NDIS_STATUS NDIS_API
SLAN_Send(
    IN NDIS_HANDLE Context,
    IN PNDIS_PACKET Packet
    )
{
	DBG_FUNC("SLAN_Send")
	PSLAN_BINDING pbinding = (PSLAN_BINDING) Context;
	PSLAN_ADAPTER adapter = pbinding->Adapter;
	PNDIS_BUFFER current_buffer;
	PVOID current_block;
	UINT current_block_length, numbuffers;
	UINT totallen;
	PVOID blockptr;
	NDIS_STATUS status;

	DBG_ENTER();

	if(adapter->HasLink == FALSE)
	{
		DBG_LEAVE(NDIS_STATUS_SUCCESS);
		//return NDIS_STATUS_SUCCESS;	
		return NDIS_STATUS_NO_CABLE;
	}

	ASSERT(adapter->ScratchBuffer);

	NdisQueryPacket(Packet, NULL, &numbuffers, &current_buffer, NULL);

	/* what an annoying little bug. */
	/* if this packet actually has no buffers, we need to pretend it never
	 * happened. why a packet has no buffers is beyond me. */

	//DBG_PRINT(("Numbuffers: %d\n", numbuffers););

	if(numbuffers == 0)
		return NDIS_STATUS_SUCCESS;

	NdisQueryBuffer(current_buffer, &current_block, &current_block_length);

    // optimization to avoid multiple copies.. blah
	// if the buffer is not word aligned, we need to copy anyway.
	// memory we allocate is guaranteed to be word aligned.
	if(numbuffers == 1 && 
		!((DWORD) current_block & 0x03))
	{
	   if(current_block_length)
	      status = deal_with_sent_buffer(adapter, current_block, current_block_length);
	   else
		  status = NDIS_STATUS_SUCCESS; // zero bytes, how nice of the protocol
	
	   DBG_LEAVE(status);
   	   return status;
	}

	blockptr = adapter->ScratchBuffer;
	totallen = 0;

	// working_block is now a pointer to data that is working_len long!
	// this is part of our packet!

	while(1)
	{
		PNDIS_BUFFER tmpbuffer;

		//DBG_PRINT(("Len %d\n", current_block_length););

		if(totallen + current_block_length > SCRATCH_BUFFER_LENGTH)
		{
			DBG_PRINT(("Buffer is too big to fit in scratch buffer, packet dropped.\n"););
			DBG_LEAVE(-1);
			// I'm dropping the packet here as a 'success'
			// because it would seem that saying "failed" could
			// cause an infinite loop as it tried again, thinking
			// it was some kind of transient adapter error?
			// anyway, this should never really happen, unless something
			// is horribly broken, we should NEVER get a packet
			// larger than our max frame (1514 - ethhdr - overhead)
			return NDIS_STATUS_SUCCESS;
		}

		NdisMoveMemory(blockptr, current_block, current_block_length);
		blockptr = (PUCHAR) blockptr + current_block_length;
		totallen += current_block_length;

		// move on to the next buffer
		NdisGetNextBuffer(current_buffer, &tmpbuffer);
		if (tmpbuffer == NULL) // no more buffers
			break;
		current_buffer = tmpbuffer;
		NdisQueryBuffer(current_buffer, &current_block, &current_block_length);
	}

	DBG_PRINT(("Buffer length: %d\n", totallen););

	status = deal_with_sent_buffer(adapter, adapter->ScratchBuffer, totallen);
	
	DBG_LEAVE(status);
	return status;
}

NDIS_STATUS NDIS_API
SLAN_TransferData(
    IN NDIS_HANDLE MacBindingHandle,
    IN NDIS_HANDLE MacReceiveContext,
    IN UINT ByteOffset,              
    IN UINT BytesToTransfer,         
    OUT PNDIS_PACKET Packet,         
    OUT PUINT BytesTransferred       
    )
{
	DBG_FUNC("SLAN_TransferData")

	// technically, this should never be called. I hope.
	// and it seems like nothing calls it, for now.
	// According to the documentation, nothing should call this, as long as
	// we pass the full packet in NdisIndicateReceive().
	// but, who knows if that is actually true.
		
	DBG_ENTER();
	DBG_LEAVE(-1);
	return NDIS_STATUS_FAILURE;
}


// this table is for looking up OIDS and sending them to the
// debug log. (ripped from some driver, somewhere).
// not too useful.

#if DBG_OIDQUERIES
static struct _SupportedOidArray {

	PCHAR		OidName;
	NDIS_OID	Oid;

} SupportedOidArray[ ] = {

	"OID_GEN_SUPPORTED_LIST",				0x00010101,
	"OID_GEN_HARDWARE_STATUS",				0x00010102,
	"OID_GEN_MEDIA_SUPPORTED",				0x00010103,
	"OID_GEN_MEDIA_IN_USE",					0x00010104,
	"OID_GEN_MAXIMUM_LOOKAHEAD",			0x00010105,
	"OID_GEN_MAXIMUM_FRAME_SIZE",			0x00010106,
	"OID_GEN_LINK_SPEED",					0x00010107,
	"OID_GEN_TRANSMIT_BUFFER_SPACE",		0x00010108,
	"OID_GEN_RECEIVE_BUFFER_SPACE",			0x00010109,
	"OID_GEN_TRANSMIT_BLOCK_SIZE",			0x0001010A,
	"OID_GEN_RECEIVE_BLOCK_SIZE",			0x0001010B,
	"OID_GEN_VENDOR_ID",					0x0001010C,
	"OID_GEN_VENDOR_DESCRIPTION",			0x0001010D,
	"OID_GEN_CURRENT_PACKET_FILTER",		0x0001010E,
	"OID_GEN_CURRENT_LOOKAHEAD",			0x0001010F,
	"OID_GEN_DRIVER_VERSION",				0x00010110,
	"OID_GEN_MAXIMUM_TOTAL_SIZE",			0x00010111,
	"OID_GEN_PROTOCOL_OPTIONS",				0x00010112,
	"OID_GEN_MAC_OPTIONS",					0x00010113,
	"OID_GEN_MEDIA_CONNECT_STATUS",			0x00010114,
	"OID_GEN_XMIT_OK",						0x00020101,
	"OID_GEN_RCV_OK",						0x00020102,
	"OID_GEN_XMIT_ERROR",					0x00020103,
	"OID_GEN_RCV_ERROR",					0x00020104,
	"OID_GEN_RCV_NO_BUFFER",				0x00020105,
	"OID_GEN_DIRECTED_BYTES_XMIT",			0x00020201,
	"OID_GEN_DIRECTED_FRAMES_XMIT",			0x00020202,
	"OID_GEN_MULTICAST_BYTES_XMIT",			0x00020203,
	"OID_GEN_MULTICAST_FRAMES_XMIT",		0x00020204,
	"OID_GEN_BROADCAST_BYTES_XMIT",			0x00020205,
	"OID_GEN_BROADCAST_FRAMES_XMIT",		0x00020206,
	"OID_GEN_DIRECTED_BYTES_RCV",			0x00020207,
	"OID_GEN_DIRECTED_FRAMES_RCV",			0x00020208,
	"OID_GEN_MULTICAST_BYTES_RCV",			0x00020209,
	"OID_GEN_MULTICAST_FRAMES_RCV",			0x0002020A,
	"OID_GEN_BROADCAST_BYTES_RCV",			0x0002020B,
	"OID_GEN_BROADCAST_FRAMES_RCV",			0x0002020C,
	"OID_GEN_RCV_CRC_ERROR",				0x0002020D,
	"OID_GEN_TRANSMIT_QUEUE_LENGTH",		0x0002020E,
	"OID_802_3_PERMANENT_ADDRESS",			0x01010101,
	"OID_802_3_CURRENT_ADDRESS",			0x01010102,
	"OID_802_3_MULTICAST_LIST",				0x01010103,
	"OID_802_3_MAXIMUM_LIST_SIZE",			0x01010104,
	"OID_802_3_RCV_ERROR_ALIGNMENT",		0x01020101,
	"OID_802_3_XMIT_ONE_COLLISION",			0x01020102,
	"OID_802_3_XMIT_MORE_COLLISIONS",		0x01020103,
	"OID_802_3_XMIT_DEFERRED",				0x01020201,
	"OID_802_3_XMIT_MAX_COLLISIONS",		0x01020202,
	"OID_802_3_RCV_OVERRUN",				0x01020203,
	"OID_802_3_XMIT_UNDERRUN",				0x01020204,
	"OID_802_3_XMIT_HEARTBEAT_FAILURE",		0x01020205,
	"OID_802_3_XMIT_TIMES_CRS_LOST",		0x01020206,
	"OID_802_3_XMIT_LATE_COLLISIONS",		0x01020207,
	"OID_802_3_PRIORITY",					0x01020208,
	"OID_802_5_PERMANENT_ADDRESS",			0x02010101,
	"OID_802_5_CURRENT_ADDRESS",			0x02010102,
	"OID_802_5_CURRENT_FUNCTIONAL",			0x02010103,
	"OID_802_5_CURRENT_GROUP",				0x02010104,
	"OID_802_5_LAST_OPEN_STATUS",			0x02010105,
	"OID_802_5_CURRENT_RING_STATUS",		0x02010106,
	"OID_802_5_CURRENT_RING_STATE",			0x02010107,
	"OID_802_5_LINE_ERRORS",				0x02020101,
	"OID_802_5_LOST_FRAMES",				0x02020102,
	"OID_802_5_BURST_ERRORS",				0x02020201,
	"OID_802_5_AC_ERRORS",					0x02020202,
	"OID_802_5_ABORT_DELIMETERS",			0x02020203,
	"OID_802_5_FRAME_COPIED_ERRORS",		0x02020204,
	"OID_802_5_FREQUENCY_ERRORS",			0x02020205,
	"OID_802_5_TOKEN_ERRORS",				0x02020206,
	"OID_802_5_INTERNAL_ERRORS",			0x02020207,
	"OID_FDDI_LONG_PERMANENT_ADDR",			0x03010101,
	"OID_FDDI_LONG_CURRENT_ADDR",			0x03010102,
	"OID_FDDI_LONG_MULTICAST_LIST",			0x03010103,
	"OID_FDDI_LONG_MAX_LIST_SIZE",			0x03010104,
	"OID_FDDI_SHORT_PERMANENT_ADDR",		0x03010105,
	"OID_FDDI_SHORT_CURRENT_ADDR",			0x03010106,
	"OID_FDDI_SHORT_MULTICAST_LIST",		0x03010107,
	"OID_FDDI_SHORT_MAX_LIST_SIZE",			0x03010108,
	"OID_FDDI_ATTACHMENT_TYPE",				0x03020101,
	"OID_FDDI_UPSTREAM_NODE_LONG",			0x03020102,
	"OID_FDDI_DOWNSTREAM_NODE_LONG",		0x03020103,
	"OID_FDDI_FRAME_ERRORS",				0x03020104,
	"OID_FDDI_FRAMES_LOST",					0x03020105,
	"OID_FDDI_RING_MGT_STATE",				0x03020106,
	"OID_FDDI_LCT_FAILURES",				0x03020107,
	"OID_FDDI_LEM_REJECTS",					0x03020108,
	"OID_FDDI_LCONNECTION_STATE",			0x03020109,
	"OID_FDDI_SMT_STATION_ID",				0x03030201,
	"OID_FDDI_SMT_OP_VERSION_ID",			0x03030202,
	"OID_FDDI_SMT_HI_VERSION_ID",			0x03030203,
	"OID_FDDI_SMT_LO_VERSION_ID",			0x03030204,
	"OID_FDDI_SMT_MANUFACTURER_DATA",		0x03030205,
	"OID_FDDI_SMT_USER_DATA",				0x03030206,
	"OID_FDDI_SMT_MIB_VERSION_ID",			0x03030207,
	"OID_FDDI_SMT_MAC_CT",					0x03030208,
	"OID_FDDI_SMT_NON_MASTER_CT",			0x03030209,
	"OID_FDDI_SMT_MASTER_CT",				0x0303020A,
	"OID_FDDI_SMT_AVAILABLE_PATHS",			0x0303020B,
	"OID_FDDI_SMT_CONFIG_CAPABILITIES",		0x0303020C,
	"OID_FDDI_SMT_CONFIG_POLICY",			0x0303020D,
	"OID_FDDI_SMT_CONNECTION_POLICY",		0x0303020E,
	"OID_FDDI_SMT_T_NOTIFY",				0x0303020F,
	"OID_FDDI_SMT_STAT_RPT_POLICY",			0x03030210,
	"OID_FDDI_SMT_TRACE_MAX_EXPIRATION",	0x03030211,
	"OID_FDDI_SMT_PORT_INDEXES",			0x03030212,
	"OID_FDDI_SMT_MAC_INDEXES",				0x03030213,
	"OID_FDDI_SMT_BYPASS_PRESENT",			0x03030214,
	"OID_FDDI_SMT_ECM_STATE",				0x03030215,
	"OID_FDDI_SMT_CF_STATE",				0x03030216,
	"OID_FDDI_SMT_HOLD_STATE",				0x03030217,
	"OID_FDDI_SMT_REMOTE_DISCONNECT_FLAG",	0x03030218,
	"OID_FDDI_SMT_STATION_STATUS",			0x03030219,
	"OID_FDDI_SMT_PEER_WRAP_FLAG",			0x0303021A,
	"OID_FDDI_SMT_MSG_TIME_STAMP",			0x0303021B,
	"OID_FDDI_SMT_TRANSITION_TIME_STAMP",	0x0303021C,
	"OID_FDDI_SMT_SET_COUNT",				0x0303021D,
	"OID_FDDI_SMT_LAST_SET_STATION_ID",		0x0303021E,
	"OID_FDDI_MAC_FRAME_STATUS_FUNCTIONS",	0x0303021F,
	"OID_FDDI_MAC_BRIDGE_FUNCTIONS",		0x03030220,
	"OID_FDDI_MAC_T_MAX_CAPABILITY",		0x03030221,
	"OID_FDDI_MAC_TVX_CAPABILITY",			0x03030222,
	"OID_FDDI_MAC_AVAILABLE_PATHS",			0x03030223,
	"OID_FDDI_MAC_CURRENT_PATH",			0x03030224,
	"OID_FDDI_MAC_UPSTREAM_NBR",			0x03030225,
	"OID_FDDI_MAC_DOWNSTREAM_NBR",			0x03030226,
	"OID_FDDI_MAC_OLD_UPSTREAM_NBR",		0x03030227,
	"OID_FDDI_MAC_OLD_DOWNSTREAM_NBR",		0x03030228,
	"OID_FDDI_MAC_DUP_ADDRESS_TEST",		0x03030229,
	"OID_FDDI_MAC_REQUESTED_PATHS",			0x0303022A,
	"OID_FDDI_MAC_DOWNSTREAM_PORT_TYPE",	0x0303022B,
	"OID_FDDI_MAC_INDEX",					0x0303022C,
	"OID_FDDI_MAC_SMT_ADDRESS",				0x0303022D,
	"OID_FDDI_MAC_LONG_GRP_ADDRESS",		0x0303022E,
	"OID_FDDI_MAC_SHORT_GRP_ADDRESS",		0x0303022F,
	"OID_FDDI_MAC_T_REQ",					0x03030230,
	"OID_FDDI_MAC_T_NEG",					0x03030231,
	"OID_FDDI_MAC_T_MAX",					0x03030232,
	"OID_FDDI_MAC_TVX_VALUE",				0x03030233,
	"OID_FDDI_MAC_T_PRI0",					0x03030234,
	"OID_FDDI_MAC_T_PRI1",					0x03030235,
	"OID_FDDI_MAC_T_PRI2",					0x03030236,
	"OID_FDDI_MAC_T_PRI3",					0x03030237,
	"OID_FDDI_MAC_T_PRI4",					0x03030238,
	"OID_FDDI_MAC_T_PRI5",					0x03030239,
	"OID_FDDI_MAC_T_PRI6",					0x0303023A,
	"OID_FDDI_MAC_FRAME_CT",				0x0303023B,
	"OID_FDDI_MAC_COPIED_CT",				0x0303023C,
	"OID_FDDI_MAC_TRANSMIT_CT",				0x0303023D,
	"OID_FDDI_MAC_TOKEN_CT",				0x0303023E,
	"OID_FDDI_MAC_ERROR_CT",				0x0303023F,
	"OID_FDDI_MAC_LOST_CT",					0x03030240,
	"OID_FDDI_MAC_TVX_EXPIRED_CT",			0x03030241,
	"OID_FDDI_MAC_NOT_COPIED_CT",			0x03030242,
	"OID_FDDI_MAC_LATE_CT",					0x03030243,
	"OID_FDDI_MAC_RING_OP_CT",				0x03030244,
	"OID_FDDI_MAC_FRAME_ERROR_THRESHOLD",	0x03030245,
	"OID_FDDI_MAC_FRAME_ERROR_RATIO",		0x03030246,
	"OID_FDDI_MAC_NOT_COPIED_THRESHOLD",	0x03030247,
	"OID_FDDI_MAC_NOT_COPIED_RATIO",		0x03030248,
	"OID_FDDI_MAC_RMT_STATE",				0x03030249,
	"OID_FDDI_MAC_DA_FLAG",					0x0303024A,
	"OID_FDDI_MAC_UNDA_FLAG",				0x0303024B,
	"OID_FDDI_MAC_FRAME_ERROR_FLAG",		0x0303024C,
	"OID_FDDI_MAC_NOT_COPIED_FLAG",			0x0303024D,
	"OID_FDDI_MAC_MA_UNITDATA_AVAILABLE",	0x0303024E,
	"OID_FDDI_MAC_HARDWARE_PRESENT",		0x0303024F,
	"OID_FDDI_MAC_MA_UNITDATA_ENABLE",		0x03030250,
	"OID_FDDI_PATH_INDEX",					0x03030251,
	"OID_FDDI_PATH_RING_LATENCY",			0x03030252,
	"OID_FDDI_PATH_TRACE_STATUS",			0x03030253,
	"OID_FDDI_PATH_SBA_PAYLOAD",			0x03030254,
	"OID_FDDI_PATH_SBA_OVERHEAD",			0x03030255,
	"OID_FDDI_PATH_CONFIGURATION",			0x03030256,
	"OID_FDDI_PATH_T_R_MODE",				0x03030257,
	"OID_FDDI_PATH_SBA_AVAILABLE",			0x03030258,
	"OID_FDDI_PATH_TVX_LOWER_BOUND",		0x03030259,
	"OID_FDDI_PATH_T_MAX_LOWER_BOUND",		0x0303025A,
	"OID_FDDI_PATH_MAX_T_REQ",				0x0303025B,
	"OID_FDDI_PORT_MY_TYPE",				0x0303025C,
	"OID_FDDI_PORT_NEIGHBOR_TYPE",			0x0303025D,
	"OID_FDDI_PORT_CONNECTION_POLICIES",	0x0303025E,
	"OID_FDDI_PORT_MAC_INDICATED",			0x0303025F,
	"OID_FDDI_PORT_CURRENT_PATH",			0x03030260,
	"OID_FDDI_PORT_REQUESTED_PATHS",		0x03030261,
	"OID_FDDI_PORT_MAC_PLACEMENT",			0x03030262,
	"OID_FDDI_PORT_AVAILABLE_PATHS",		0x03030263,
	"OID_FDDI_PORT_MAC_LOOP_TIME",			0x03030264,
	"OID_FDDI_PORT_PMD_CLASS",				0x03030265,
	"OID_FDDI_PORT_CONNECTION_CAPABILITIES", 0x03030266,
	"OID_FDDI_PORT_INDEX",					0x03030267,
	"OID_FDDI_PORT_MAINT_LS",				0x03030268,
	"OID_FDDI_PORT_BS_FLAG",				0x03030269,
	"OID_FDDI_PORT_PC_LS",					0x0303026A,
	"OID_FDDI_PORT_EB_ERROR_CT",			0x0303026B,
	"OID_FDDI_PORT_LCT_FAIL_CT",			0x0303026C,
	"OID_FDDI_PORT_LER_ESTIMATE",			0x0303026D,
	"OID_FDDI_PORT_LEM_REJECT_CT",			0x0303026E,
	"OID_FDDI_PORT_LEM_CT",					0x0303026F,
	"OID_FDDI_PORT_LER_CUTOFF",				0x03030270,
	"OID_FDDI_PORT_LER_ALARM",				0x03030271,
	"OID_FDDI_PORT_CONNNECT_STATE",			0x03030272,
	"OID_FDDI_PORT_PCM_STATE",				0x03030273,
	"OID_FDDI_PORT_PC_WITHHOLD",			0x03030274,
	"OID_FDDI_PORT_LER_FLAG",				0x03030275,
	"OID_FDDI_PORT_HARDWARE_PRESENT",		0x03030276,
	"OID_FDDI_SMT_STATION_ACTION",			0x03030277,
	"OID_FDDI_PORT_ACTION",					0x03030278,
	"OID_FDDI_IF_DESCR",					0x03030279,
	"OID_FDDI_IF_TYPE",						0x0303027A,
	"OID_FDDI_IF_MTU",						0x0303027B,
	"OID_FDDI_IF_SPEED",					0x0303027C,
	"OID_FDDI_IF_PHYS_ADDRESS",				0x0303027D,
	"OID_FDDI_IF_ADMIN_STATUS",				0x0303027E,
	"OID_FDDI_IF_OPER_STATUS",				0x0303027F,
	"OID_FDDI_IF_LAST_CHANGE",				0x03030280,
	"OID_FDDI_IF_IN_OCTETS",				0x03030281,
	"OID_FDDI_IF_IN_UCAST_PKTS",			0x03030282,
	"OID_FDDI_IF_IN_NUCAST_PKTS",			0x03030283,
	"OID_FDDI_IF_IN_DISCARDS",				0x03030284,
	"OID_FDDI_IF_IN_ERRORS",				0x03030285,
	"OID_FDDI_IF_IN_UNKNOWN_PROTOS",		0x03030286,
	"OID_FDDI_IF_OUT_OCTETS",				0x03030287,
	"OID_FDDI_IF_OUT_UCAST_PKTS",			0x03030288,
	"OID_FDDI_IF_OUT_NUCAST_PKTS",			0x03030289,
	"OID_FDDI_IF_OUT_DISCARDS",				0x0303028A,
	"OID_FDDI_IF_OUT_ERRORS",				0x0303028B,
	"OID_FDDI_IF_OUT_QLEN",					0x0303028C,
	"OID_FDDI_IF_SPECIFIC",					0x0303028D,
	"OID_WAN_PERMANENT_ADDRESS",			0x04010101,
	"OID_WAN_CURRENT_ADDRESS",				0x04010102,
	"OID_WAN_QUALITY_OF_SERVICE",			0x04010103,
	"OID_WAN_PROTOCOL_TYPE",				0x04010104,
	"OID_WAN_MEDIUM_SUBTYPE",				0x04010105,
	"OID_WAN_HEADER_FORMAT",				0x04010106,
	"OID_WAN_GET_INFO",						0x04010107,
	"OID_WAN_SET_LINK_INFO",				0x04010108,
	"OID_WAN_GET_LINK_INFO",				0x04010109,
	"OID_WAN_LINE_COUNT",					0x0401010A,
	"OID_WAN_GET_BRIDGE_INFO",				0x0401020A,
	"OID_WAN_SET_BRIDGE_INFO",				0x0401020B,
	"OID_WAN_GET_COMP_INFO",				0x0401020C,
	"OID_WAN_SET_COMP_INFO",				0x0401020D,
	"OID_WAN_GET_STATS_INFO",				0x0401020E,
	"OID_GET_SEND_PACKETS",					0xFF020201,
	"OID_GET_RECEIVE_PACKETS",				0xFF020202
};


#define NUM_OID_ENTRIES (sizeof(SupportedOidArray) / sizeof(SupportedOidArray[0]))

/*
// This debug routine will lookup the printable name for the selected OID.
*/
char *
GetOidString(NDIS_OID Oid)
{
    UINT i;

    for (i = 0; i < NUM_OID_ENTRIES-1; i++)
    {
		if (SupportedOidArray[i].Oid == Oid) {
			break;
		}
    }
    return(SupportedOidArray[i].OidName);
}

void NdisRequestTrace(PNDIS_REQUEST NdisRequest)
{
	DBG_FUNC("NdisRequestTrace")
	DBG_PRINT(("NdisRequest: %s\n", GetOidString(NdisRequest->DATA.QUERY_INFORMATION.Oid)););
}
#else
void NdisRequestTrace(PNDIS_REQUEST NdisRequest)
{
}
#endif //DBG_OIDQUERIES


static UINT SLAN_SupportedOids[] = {
    OID_GEN_SUPPORTED_LIST,
    OID_GEN_HARDWARE_STATUS,
    OID_GEN_MEDIA_SUPPORTED,
    OID_GEN_MEDIA_IN_USE,
    OID_GEN_MAXIMUM_FRAME_SIZE,
    OID_GEN_MAXIMUM_TOTAL_SIZE,
    OID_GEN_MAC_OPTIONS,
    OID_GEN_PROTOCOL_OPTIONS,
    OID_GEN_LINK_SPEED,
    OID_GEN_TRANSMIT_BUFFER_SPACE,
    OID_GEN_RECEIVE_BUFFER_SPACE,
    OID_GEN_TRANSMIT_BLOCK_SIZE,
    OID_GEN_RECEIVE_BLOCK_SIZE,
    OID_GEN_VENDOR_DESCRIPTION,
    OID_GEN_VENDOR_ID,
    OID_GEN_DRIVER_VERSION,
    OID_GEN_CURRENT_PACKET_FILTER,
    OID_802_3_PERMANENT_ADDRESS,
    OID_802_3_CURRENT_ADDRESS,
    OID_802_3_MULTICAST_LIST,
    OID_802_3_MAXIMUM_LIST_SIZE,
    };


NDIS_STATUS 
  SLAN_QueryInfo(
    IN NDIS_HANDLE  MiniportAdapterContext,
    IN NDIS_OID  Oid,
    IN PVOID  InformationBuffer,
    IN ULONG  InformationBufferLength,
    OUT PULONG  BytesWritten,
    OUT PULONG  BytesNeeded
    )
{
	DBG_FUNC("SLAN_QueryInfo")
	PSLAN_ADAPTER adapter = (PSLAN_ADAPTER) MiniportAdapterContext;
	NDIS_STATUS status = NDIS_STATUS_SUCCESS;
    UINT bytesleft = InformationBufferLength;
    PUCHAR infobuffer = (PUCHAR)(InformationBuffer);
    NDIS_HARDWARE_STATUS hardwarestatus = NdisHardwareStatusReady;
    NDIS_MEDIUM medium = NdisMedium802_3;
    ULONG generic_ulong;
    USHORT generic_ushort;
    UCHAR generic_array[6] = {0x00, 0x00, 0xde, 0xad, 0xbe, 0xef};
    UINT move_bytes = sizeof(ULONG);
    PVOID move_source = (PVOID)(&generic_ulong);

	DBG_ENTER();

    //
    // Make sure that int is 4 bytes.  Else GenericULong must change
    // to something of size 4.
    //
    ASSERT(sizeof(ULONG) == 4);

    //
    // Switch on request type
    //

    switch (Oid) 
	{

		case OID_GEN_MAC_OPTIONS:
	
		    generic_ulong = (ULONG)(NDIS_MAC_OPTION_TRANSFERS_NOT_PEND |
                               NDIS_MAC_OPTION_NO_LOOPBACK);

        break;

		case OID_GEN_SUPPORTED_LIST:

			move_source = (PVOID)(SLAN_SupportedOids);
			move_bytes = sizeof(SLAN_SupportedOids);
			break;

		case OID_GEN_HARDWARE_STATUS:

			hardwarestatus = NdisHardwareStatusReady;
			move_source = (PVOID)(&hardwarestatus);
			move_bytes = sizeof(NDIS_HARDWARE_STATUS);
			break;

		case OID_GEN_MEDIA_SUPPORTED:
		case OID_GEN_MEDIA_IN_USE:
			move_source = (PVOID) (&medium);
			move_bytes = sizeof(NDIS_MEDIUM);
			break;

		case OID_GEN_MAXIMUM_FRAME_SIZE:
			generic_ulong = (ULONG)(1514 - (14 + 160));
			break;

		case OID_GEN_MAXIMUM_TOTAL_SIZE:
			generic_ulong = (ULONG)(1514);
			break;

		case OID_GEN_LINK_SPEED:
			generic_ulong = (ULONG)(1000000);
			break;

		case OID_GEN_TRANSMIT_BUFFER_SPACE:
			generic_ulong = 10240;
			break;

	    case OID_GEN_RECEIVE_BUFFER_SPACE:
			generic_ulong = 10240;
			break;

	    case OID_GEN_TRANSMIT_BLOCK_SIZE:
			generic_ulong = (ULONG)(256);
			break;

		case OID_GEN_RECEIVE_BLOCK_SIZE:
			generic_ulong = (ULONG)(256);
			break;

	    case OID_GEN_VENDOR_ID:
			generic_ulong = 0x00;
	        break;

		case OID_GEN_VENDOR_DESCRIPTION:
			move_source = (PVOID)"Lightlink SLAN Adapter";
			move_bytes = 23;
			break;

	    case OID_GEN_DRIVER_VERSION:

			generic_ushort = ((USHORT) 3 << 8) | 0;
			move_source = (PVOID)(&generic_ushort);
			move_bytes = sizeof(generic_ushort);
			break;

	    case OID_802_3_PERMANENT_ADDRESS:
			move_source = (PVOID)(generic_array);
			move_bytes = 6;
			break;

		case OID_802_3_CURRENT_ADDRESS:
			move_source = (PVOID)(generic_array);
			move_bytes = 6;
			break;
        
		case OID_802_3_MAXIMUM_LIST_SIZE:
			generic_ulong = 8; // sure, why not? ;)
			break;

		default:
			status = NDIS_STATUS_INVALID_OID;
			break;
    }


    if (status == NDIS_STATUS_SUCCESS) 
	{
        if (move_bytes > bytesleft) 
		{
            // Not enough room in InformationBuffer!
            *BytesNeeded = move_bytes;
            status = NDIS_STATUS_INVALID_LENGTH;
        } 
		else 
		{
			NdisMoveMemory(infobuffer, move_source, move_bytes);
            (*BytesWritten) = move_bytes;
        }
    }

	DBG_LEAVE(status);
    return status;
}

NDIS_STATUS NDIS_API
SLAN_QueryGlobalStatistics(
    IN NDIS_HANDLE MacContext,
    IN PNDIS_REQUEST NdisRequest
    )
{
	DBG_FUNC("SLAN_QueryGlobalStatistics")
	PSLAN_ADAPTER adapter = (PSLAN_ADAPTER) MacContext;
	NDIS_STATUS status;
    UINT BytesWritten = 0;
    UINT BytesNeeded = 0;
    UINT BytesLeft = NdisRequest->DATA.QUERY_INFORMATION.InformationBufferLength;
    PUCHAR InfoBuffer = (PUCHAR)(NdisRequest->DATA.QUERY_INFORMATION.InformationBuffer);

	DBG_ENTER()

	status = SLAN_QueryInfo((NDIS_HANDLE) adapter, NdisRequest->DATA.QUERY_INFORMATION.Oid,
							InfoBuffer, BytesLeft,
							&BytesWritten, &BytesNeeded);

    NdisRequest->DATA.QUERY_INFORMATION.BytesWritten = BytesWritten;
    NdisRequest->DATA.QUERY_INFORMATION.BytesNeeded = BytesNeeded;

	DBG_LEAVE(status);
	return status;
}

NDIS_STATUS NDIS_API
SLAN_Request(
    IN NDIS_HANDLE MacBindingHandle,
    IN PNDIS_REQUEST NdisRequest
    )
{
	DBG_FUNC("SLAN_Request")
	NDIS_STATUS status = NDIS_STATUS_SUCCESS;

	DBG_ENTER()
	if(NdisRequest->RequestType == NdisRequestQueryInformation)
	{
		DBG_PRINT(("NdisRequestQueryInformation :: \n"););
		status = SLAN_QueryGlobalStatistics(MacBindingHandle, NdisRequest);
	}
	else if(NdisRequest->RequestType == NdisRequestSetInformation)
	{
		// suuure, we set this in our adapter.. :)
        status = NDIS_STATUS_SUCCESS;

		DBG_PRINT(("NdisRequestSetInformation :: \n"););

        NdisRequest->DATA.SET_INFORMATION.BytesRead = 0;
        NdisRequest->DATA.SET_INFORMATION.BytesNeeded = 0;
	}
	else
	{
		DBG_PRINT(("NdisRequest UNKNOWN!!!!! :: \n"););
		status = NDIS_STATUS_NOT_RECOGNIZED;
	}

	NdisRequestTrace(NdisRequest);

	DBG_LEAVE(status);
	return status;
}
