/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
  $Header: /home/cvsroot/stage4/platform/windows/installer/installer.cpp,v 1.1 2001/07/06 17:13:45 lucas Exp $
  $Author: lucas $ $Date: 2001/07/06 17:13:45 $ $Name:  $
  $Revision: 1.1 $
 
*/

#include "stdafx.h"

#define MSTCP_LOCAL "Enum\\Network\\MSTCP"
#define VREDIR_LOCAL "Enum\\Network\\VREDIR"

LPVOID GetLastErrorMessage(int errnum)
{
	LPVOID lpMsgBuf;

	FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | 
		FORMAT_MESSAGE_FROM_SYSTEM | 
		FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		errnum,
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
		(LPTSTR) &lpMsgBuf,
		0,
		NULL);

	return lpMsgBuf;
}

#define RET_ERROR (-1)

void RecursiveCopyKeys(HKEY srckey, HKEY dstkey)
{
	TCHAR name[1024];
	DWORD namelen;
	HKEY nsrckey, ndstkey;
	int i;

	printf("messing with registry, this may take a bit...\n");

	for(i = 0; i < 9999; i++)
	{
		FILETIME ft;

		namelen = 1024;
		if(RegEnumKeyEx(srckey, i, name, &namelen, NULL, NULL, NULL, &ft) != ERROR_SUCCESS)
			break;

		RegOpenKeyEx(srckey, name, 0, NULL, &nsrckey);
		RegCreateKeyEx(dstkey, name, 0, "", REG_OPTION_NON_VOLATILE, NULL, NULL, &ndstkey, NULL);
		RecursiveCopyKeys(nsrckey, ndstkey);
		RegCloseKey(nsrckey);
		RegCloseKey(ndstkey);
	}

	for(i = 0; i < 9999; i++)
	{
		char dataname[1024];
		BYTE keydata[1024];
		ULONG datalen, datanamelen;
		DWORD keytype;
		char scratch[1024];

		datanamelen = datalen = 1024;
		
		if(RegEnumValue(srckey, i, dataname, &datanamelen, NULL, &keytype, keydata, &datalen) != ERROR_SUCCESS)
			break;

		if(stricmp(dataname, "NodeType") == 0)
			continue;

		if(stricmp(dataname, "IPAddress") == 0)
		{
			datalen = sprintf(scratch, "0.0.0.0");
			datalen++;

			memcpy(keydata, scratch, 1024);
		}
		else if(stricmp(dataname, "IPMask") == 0)
		{
			datalen = sprintf(scratch, "0.0.0.0");
			datalen++;

			memcpy(keydata, scratch, 1024);
		}

		RegSetValueEx(dstkey, dataname, 0, keytype, keydata, datalen);
	}
}

int CloneNew(char *keyname)
{
	HKEY srckey, dstkey;
	char *service, *id;
	BOOL foundmyidx = FALSE;
	int myidx = -1, index;
	TCHAR name[1024];
	long errval;
	char fullkey[1024], newkey[1024];

	printf("Clonenew(\"%s\")\n", keyname);

	id = strrchr(keyname, '\\');
	if(!id)
		return RET_ERROR;

	*id = '\0';
	id++;
	service = keyname;

	_snprintf(fullkey, 1024, "System\\CurrentControlSet\\Services\\Class\\%s", service, id);

    if((errval = RegOpenKeyEx(HKEY_LOCAL_MACHINE, fullkey, 0, KEY_READ, &srckey)) != ERROR_SUCCESS)
		return RET_ERROR;

	while(!foundmyidx)
	{
		myidx++;

		for(index = 0; index < 9999; index++)
		{
			FILETIME ft;
			ULONG namelen;
			int val;

			namelen = 1024;
			if((errval = RegEnumKeyEx(srckey, index, name, &namelen, NULL, NULL, NULL, &ft)) != ERROR_SUCCESS)
			{
				if(index == 0 && errval == ERROR_NO_MORE_ITEMS)
				{
					MessageBox(NULL, "Could not duplicate, no entry found to duplicate!", "Error duplicating mstcp", MB_OK|MB_ICONSTOP);
					return RET_ERROR;
				}

				if(errval != ERROR_NO_MORE_ITEMS)
					return RET_ERROR;

				foundmyidx = TRUE;
				break;
			}

			val = strtol(name, NULL, 10);
			if(val == myidx)
				break;
		}
	}

	RegCloseKey(srckey);

	_snprintf(fullkey, 1024, "System\\CurrentControlSet\\Services\\Class\\%s\\%s", service, id);

    if((errval = RegOpenKeyEx(HKEY_LOCAL_MACHINE, fullkey, 0, KEY_READ, &srckey)) != ERROR_SUCCESS)
		return RET_ERROR;

	_snprintf(newkey, 1024, "System\\CurrentControlSet\\Services\\Class\\%s\\%04d", service, myidx);

	RegCreateKeyEx(HKEY_LOCAL_MACHINE, newkey, 0, "", REG_OPTION_NON_VOLATILE, NULL, NULL, &dstkey, NULL);

	RecursiveCopyKeys(srckey, dstkey);

	RegCloseKey(srckey);
	RegCloseKey(dstkey);

	return myidx;
}

#define MAXBINDINGS 40

int EnumCopyObject(char *basekey, char *keyname, char *subkeyname)
{
	HKEY mainkey, srckey, bindingskey, newkey;
	int myidx = -1, index;
	BOOL foundmyidx = FALSE;
	long errval;
	LPCTSTR err;
	char myidxbuf[5];
	TCHAR name[1024];
	char vname[1024], newname[1024];
	char fullkey[1024];
	char *bindings[MAXBINDINGS];
	int numbindings;
	BOOL isNetTrans = FALSE;

	printf("EnumCopyObject(\"%s\", \"%s\", \"%s\")\n", basekey, keyname, subkeyname ? subkeyname : "NULL");

	_snprintf(fullkey, 1024, "%s\\%s", basekey, keyname);

	if((errval = RegOpenKeyEx(HKEY_LOCAL_MACHINE, fullkey, 0, KEY_READ, &mainkey)) != ERROR_SUCCESS)
	{
		err = (LPCTSTR) GetLastErrorMessage(errval);
		MessageBox(NULL, err, "Could not access key -- is mstcp installed?", MB_OK|MB_ICONSTOP);
		LocalFree((LPVOID) err);
		return RET_ERROR;
	}

	while(!foundmyidx)
	{
		myidx++;

		for(index = 0; index < 9999; index++)
		{
			FILETIME ft;
			ULONG namelen;
			int val;

			namelen = 1024;
			if((errval = RegEnumKeyEx(mainkey, index, name, &namelen, NULL, NULL, NULL, &ft)) != ERROR_SUCCESS)
			{
				if(index == 0 && errval == ERROR_NO_MORE_ITEMS)
				{
					MessageBox(NULL, "MSTCP is not bound to anything!", "Could not access MSTCP key.", MB_OK|MB_ICONSTOP);
					return RET_ERROR;
				}

				if(errval != ERROR_NO_MORE_ITEMS)
					return RET_ERROR;

				foundmyidx = TRUE;
				break;
			}

			val = strtol(name, NULL, 10);
			if(val == myidx)
				break;
		}
	}

	_snprintf(myidxbuf, 5, "%04d", myidx);
	if(myidx != strtol(myidxbuf, NULL, 10))
	{
		MessageBox(NULL, myidxbuf, "Error with new name buffer", MB_ICONHAND|MB_OK);
		return RET_ERROR;
	}

	printf("I choose index %s (%d)\n", myidxbuf, myidx);

	if(subkeyname)
	{
		_snprintf(vname, 1024, "%s\\%s\\Bindings", fullkey, subkeyname);
	}
	else
	{
		FILETIME ft;
		char bname[1024];
		ULONG bnamelen = 1024;

		RegEnumKeyEx(mainkey, 0, bname, &bnamelen, NULL, NULL, NULL, &ft);
		_snprintf(vname, 1024, "%s\\%s\\Bindings", fullkey, bname);
	}

	numbindings = 0;

	if(RegOpenKeyEx(HKEY_LOCAL_MACHINE, vname, 0, NULL, &bindingskey) == ERROR_SUCCESS)
	{
		char *service, *id;
		int i;

		for(i = 0; i < 9999; i++)
		{
			char dataname[1024], *x;
			BYTE keydata[1024];
			ULONG datalen, datanamelen;
			DWORD keytype;
			int rv;

			datanamelen = datalen = 1024;

			if(RegEnumValue(bindingskey, i, dataname, &datanamelen, NULL, &keytype, keydata, &datalen) != ERROR_SUCCESS)
				break;

			printf("%s bindings: %s\n", vname, dataname);

			service = dataname;
			x = strrchr(dataname, '\\');
			if(x)
			{
				*x = '\0';
				id = x+1;
				rv = EnumCopyObject("Enum\\Network", dataname, id);
				if(rv == RET_ERROR)
				{
					printf("EnumCopyObject(%s, %s) FAILED!\n", dataname, id);
					return rv;
				}
			}

			if(numbindings+1 < MAXBINDINGS)
			{
				bindings[numbindings] = (char *) malloc(datanamelen + 16);
				_snprintf(bindings[numbindings], datanamelen + 16, "%s\\%04d", service, rv);
				numbindings++;
			}
		}

		RegCloseKey(bindingskey);
	}

	/* now we need to copy the nettrans stuff.. things that "Driver" is bound to. */

	/* vname is what I am copying from */
	/* newname is what I am copying to */

	_snprintf(newname, 1024, "%s\\%s", fullkey, myidxbuf);

	if(subkeyname)
	{
		_snprintf(vname, 1024, "%s\\%s", fullkey, subkeyname);
	}
	else
	{
		FILETIME ft;
		char bname[1024];
		ULONG bnamelen = 1024;

		RegEnumKeyEx(mainkey, 0, bname, &bnamelen, NULL, NULL, NULL, &ft);
		_snprintf(vname, 1024, "%s\\%s", fullkey, bname);
	}

	if(RegOpenKeyEx(HKEY_LOCAL_MACHINE, vname, 0, NULL, &srckey) == ERROR_SUCCESS)
	{
		int driverid;
		DWORD kdatalen = 1024;
		char kdata[1024];
		DWORD ktype;
		char driveridbuf[5];
		int i;
		DWORD dispo;

		RegCreateKeyEx(HKEY_LOCAL_MACHINE, newname, 0, "", REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &newkey, &dispo);

		if(dispo == REG_OPENED_EXISTING_KEY)
		{
			RegCloseKey(srckey);
			RegCloseKey(newkey);
			MessageBox(NULL, "KEY ALREADY EXISTS!", "Bailing out!", MB_OK|MB_ICONHAND);
			return RET_ERROR;
		}

		if(RegQueryValueEx(srckey, "Driver", NULL, &ktype, (BYTE *) kdata, &kdatalen) == ERROR_SUCCESS)
		{
			char scratch[1024], *x;

			_snprintf(scratch, 1024, "%s", kdata);
			x = strrchr(scratch, '\\');
			if(!x)
				return RET_ERROR;
			*x = '\0';
			x++;

			/* 
			   if it is a protocol, clone a new one.
			   DO NOT CLONE SERVICES. this was ugly ;)   
			*/

			if(strnicmp(scratch, "NetTrans", 8) == 0)
			{
				driverid = CloneNew(kdata);
				isNetTrans = TRUE;
			}
			else
			{
				_snprintf(scratch, 1024, "%s", kdata);

				driverid = strtol(x, NULL, 10);
			}
		}
		else
			return RET_ERROR;

		_snprintf(driveridbuf, 5, "%04d", driverid);

		for(i = 0; i < 9999; i++)
		{
			char dataname[1024], *x;
			BYTE keydata[1024];
			ULONG datalen, datanamelen;
			DWORD keytype;
			char scratch[1024];

			datanamelen = datalen = 1024;

			if(RegEnumValue(srckey, i, dataname, &datanamelen, NULL, &keytype, keydata, &datalen) != ERROR_SUCCESS)
				break;

			if(stricmp(dataname, "Driver") == 0)
			{
				x = strrchr((char *)keydata, '\\');
				if(x) *x = '\0';

				datalen = _snprintf(scratch, 1024, "%s\\%s", keydata, driveridbuf);
				datalen++;
				memcpy(keydata, scratch, 1024);
			}
			else if (stricmp(dataname, "MasterCopy") == 0 && isNetTrans)
			{
				x = strrchr((char *)keydata, '\\');
				if(x) *x = '\0';

				datalen = _snprintf(scratch, 1024, "%s\\%s", keydata, myidxbuf);
				datalen++;
				memcpy(keydata, scratch, 1024);
			}

			RegSetValueEx(newkey, dataname, 0, keytype, keydata, datalen);
		}

		RegCloseKey(newkey);
		_snprintf(newname, 1024, "%s\\%s\\Bindings", fullkey, myidxbuf);
		RegCreateKeyEx(HKEY_LOCAL_MACHINE, newname, 0, "", REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &newkey, &dispo);

		for(i = 0; i < numbindings; i++)
		{
			char *tc = "";

			RegSetValueEx(newkey, bindings[i], 0, REG_SZ, (BYTE *) tc, 1);
			free(bindings[i]);
		}

		RegCloseKey(newkey);
		RegCloseKey(srckey);
	}

	RegCloseKey(mainkey);
	return myidx;
}

int CreateSlanNet()
{
	HKEY netkey;
	BOOL foundmyidx = FALSE;
	int myidx = 0, index;
	long errval;
	TCHAR name[1024];
	char newname[1024];
	UCHAR binary;

	printf("Installing SLAN net driver...\n");

	if(RegCreateKeyEx(HKEY_LOCAL_MACHINE, "System\\CurrentControlSet\\Services\\Class\\Net",
		0, "", REG_OPTION_NON_VOLATILE, NULL, NULL, &netkey, NULL) != ERROR_SUCCESS)
	{
		printf("Error creating slan net driver\n");
		return RET_ERROR;
	}

	while(!foundmyidx)
	{
		myidx++;

		for(index = 0; index < 9999; index++)
		{
			FILETIME ft;
			ULONG namelen;
			int val;

			namelen = 1024;
			if((errval = RegEnumKeyEx(netkey, index, name, &namelen, NULL, NULL, NULL, &ft)) != ERROR_SUCCESS)
			{
				if(errval != ERROR_NO_MORE_ITEMS)
					return RET_ERROR;

				foundmyidx = TRUE;
				break;
			}

			val = strtol(name, NULL, 10);
			if(val == myidx)
				break;
		}
	}

	RegCloseKey(netkey);

	_snprintf(newname, 1024, "System\\CurrentControlSet\\Services\\Class\\Net\\%04d", myidx);
	if(RegCreateKeyEx(HKEY_LOCAL_MACHINE, newname,
		0, "", REG_OPTION_NON_VOLATILE, NULL, NULL, &netkey, NULL) != ERROR_SUCCESS)
	{
		printf("Error creating slan net driver (main)\n");
		return RET_ERROR;
	}

	RegSetValueEx(netkey, "DeviceVxDs", 0, REG_SZ, (BYTE *) "SLANMAC.VXD", 12);
	RegSetValueEx(netkey, "DevLoader" , 0, REG_SZ, (BYTE *) "*ndis", 6);
	RegSetValueEx(netkey, "DeviceDesc" , 0, REG_SZ,
					(BYTE *) "Lightlink SLAN Networking Driver", 33);
	/* this should really come from the VXD... */
	RegSetValueEx(netkey, "DriverDate" , 0, REG_SZ, (BYTE *) "06-10-01", 9);
	RegSetValueEx(netkey, "EnumPropPages" , 0, REG_SZ, (BYTE *) "netdi.dll,EnumPropPages", 24);

	RegCloseKey(netkey);

	_snprintf(newname, 1024, "System\\CurrentControlSet\\Services\\Class\\Net\\%04d\\Ndi", myidx);
	if(RegCreateKeyEx(HKEY_LOCAL_MACHINE, newname,
		0, "", REG_OPTION_NON_VOLATILE, NULL, NULL, &netkey, NULL) != ERROR_SUCCESS)
	{
		printf("Error creating slan net driver (ndi)\n");
		return RET_ERROR;
	}

	RegSetValueEx(netkey, "DeviceID" , 0, REG_SZ, (BYTE *) "SLANMAC", 8);
	RegSetValueEx(netkey, "HelpText" , 0, REG_SZ, (BYTE *)
			"Lightlink SLAN VPN Driver.", 27);

	RegCloseKey(netkey);

	_snprintf(newname, 1024, "System\\CurrentControlSet\\Services\\Class\\Net\\%04d\\Ndi\\Install", myidx);
	if(RegCreateKeyEx(HKEY_LOCAL_MACHINE, newname,
		0, "", REG_OPTION_NON_VOLATILE, NULL, NULL, &netkey, NULL) != ERROR_SUCCESS)
	{
		printf("Error creating slan net driver (ndi.install)\n");
		return RET_ERROR;
	}

	RegSetValueEx(netkey, "" , 0, REG_SZ, (BYTE *) "SLANMAC.Install", 16);

	RegCloseKey(netkey);

	_snprintf(newname, 1024, "System\\CurrentControlSet\\Services\\Class\\Net\\%04d\\Ndi\\Interfaces", myidx);
	if(RegCreateKeyEx(HKEY_LOCAL_MACHINE, newname,
		0, "", REG_OPTION_NON_VOLATILE, NULL, NULL, &netkey, NULL) != ERROR_SUCCESS)
	{
		printf("Error creating slan net driver (ndi.interfaces)\n");
		return RET_ERROR;
	}

	RegSetValueEx(netkey, "DefLower" , 0, REG_SZ, (BYTE *)"ethernet", 9);
	RegSetValueEx(netkey, "DefUpper" , 0, REG_SZ, (BYTE *)"ndis3", 6);
	RegSetValueEx(netkey, "Lower" , 0, REG_SZ, (BYTE *)"ethernet", 9);
	RegSetValueEx(netkey, "LowerRange" , 0, REG_SZ, (BYTE *)"ethernet", 9);
	RegSetValueEx(netkey, "Upper" , 0, REG_SZ, (BYTE *)"ndis3", 6);
	RegSetValueEx(netkey, "UpperRange" , 0, REG_SZ, (BYTE *)"ndis3", 6);

	RegCloseKey(netkey);

	_snprintf(newname, 1024, "System\\CurrentControlSet\\Services\\Class\\Net\\%04d\\NDIS", myidx);
	if(RegCreateKeyEx(HKEY_LOCAL_MACHINE, newname,
		0, "", REG_OPTION_NON_VOLATILE, NULL, NULL, &netkey, NULL) != ERROR_SUCCESS)
	{
		printf("Error creating slan net driver (ndis)\n");
		return RET_ERROR;
	}

	RegSetValueEx(netkey, "LogDriverName" , 0, REG_SZ, (BYTE *)"SLANMAC", 8);

	binary = 3;
	RegSetValueEx(netkey, "MajorNdisVersion" , 0, REG_BINARY, &binary, 1);
	binary = 0x0a;
	RegSetValueEx(netkey, "MinorNdisVersion", 0, REG_BINARY, &binary, 1);

	RegCloseKey(netkey);

	return myidx;
}

int CreateSlanEnum(int driverid, int tcpid)
{
	HKEY netkey;
	BOOL foundmyidx = FALSE;
	int myidx = 0, index;
	long errval;
	TCHAR name[1024];
	char driverbuf[1024];
	char newname[1024];
	UCHAR binary[4];
	DWORD len;

	printf("Installing SLAN net enum driver...\n");

	if(RegCreateKeyEx(HKEY_LOCAL_MACHINE, "Enum\\Root\\SLAN",
		0, "", REG_OPTION_NON_VOLATILE, NULL, NULL, &netkey, NULL) != ERROR_SUCCESS)
	{
		printf("Error creating slan net enum driver\n");
		return RET_ERROR;
	}

	while(!foundmyidx)
	{
		myidx++;

		for(index = 0; index < 9999; index++)
		{
			FILETIME ft;
			ULONG namelen;
			int val;

			namelen = 1024;
			if((errval = RegEnumKeyEx(netkey, index, name, &namelen, NULL, NULL, NULL, &ft)) != ERROR_SUCCESS)
			{
				if(errval != ERROR_NO_MORE_ITEMS)
					return RET_ERROR;

				foundmyidx = TRUE;
				break;
			}

			val = strtol(name, NULL, 10);
			if(val == myidx)
				break;
		}
	}

	RegCloseKey(netkey);

	_snprintf(newname, 1024, "Enum\\Root\\SLAN\\%04d", myidx);
	if(RegCreateKeyEx(HKEY_LOCAL_MACHINE, newname,
		0, "", REG_OPTION_NON_VOLATILE, NULL, NULL, &netkey, NULL) != ERROR_SUCCESS)
	{
		printf("Error creating slan net driver (main)\n");
		return RET_ERROR;
	}

	RegSetValueEx(netkey, "Class" , 0, REG_SZ, (BYTE *)"Net", 4);
	RegSetValueEx(netkey, "CompatibleIDs" , 0, REG_SZ, (BYTE *)"*SLANMAC", 9);

	binary [0] = 0x10;
	binary [1] = 0x0;
	binary [2] = 0x0;
	binary [3] = 0x0;
	RegSetValueEx(netkey, "ConfigFlags", 0, REG_BINARY, *&binary, 4);
	RegSetValueEx(netkey, "DeviceDesc" , 0, REG_SZ,
				(BYTE *) "Lightlink SLAN Networking Driver", 33);

	len = _snprintf(driverbuf, 1024, "Net\\%04d", driverid);
	RegSetValueEx(netkey, "Driver" , 0, REG_SZ, (BYTE *) driverbuf, len + 1);
	RegSetValueEx(netkey, "Mfg" , 0, REG_SZ, (BYTE *)"Lightlink SLAN Team", 20);

	RegCloseKey(netkey);

	_snprintf(newname, 1024, "Enum\\Root\\SLAN\\%04d\\Bindings", myidx);
	if(RegCreateKeyEx(HKEY_LOCAL_MACHINE, newname,
		0, "", REG_OPTION_NON_VOLATILE, NULL, NULL, &netkey, NULL) != ERROR_SUCCESS)
	{
		printf("Error creating slan net driver (bindings)\n");
		return RET_ERROR;
	}

	_snprintf(driverbuf, 1024, "MSTCP\\%04d", tcpid);
	RegSetValueEx(netkey, driverbuf, 0, REG_SZ, (BYTE *) "", 1);

	RegCloseKey(netkey);

	return myidx;
}


int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{

	int tcpv, netv;
#ifndef NO_DEBUG_CONSOLE
	{
	/* This undocumented, obscure win32 voodoo magic sets up a console, remapping stderr
	and stdout... */

	int hCrt;
	FILE *hf;

	AllocConsole();

	hCrt = _open_osfhandle((long) GetStdHandle(STD_OUTPUT_HANDLE), _O_TEXT);
	hf = _fdopen( hCrt, "w" );
	*stdout = *hf;
	setvbuf(stdout, NULL, _IONBF, 0); 

	hCrt = _open_osfhandle((long) GetStdHandle(STD_ERROR_HANDLE), _O_TEXT);
	hf = _fdopen(hCrt, "w");
	*stderr = *hf;
	setvbuf(stderr, NULL, _IONBF, 0); 

	SetConsoleTitle("Install Debug");
	}
#endif


	tcpv = EnumCopyObject("Enum\\Network", "MSTCP", NULL);
	netv = CreateSlanNet();

	CreateSlanEnum(netv, tcpv);

	Sleep(1000);


	return 0;
}



