/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
  $Header: /home/cvsroot/stage4/platform/windows/slandll/slandll.c,v 1.1 2001/06/05 16:09:16 lucasm Exp $
  $Author: lucasm $ $Date: 2001/06/05 16:09:16 $ $Name:  $
  $Revision: 1.1 $
 
*/

#include "stdafx.h"
#include "slanioctl.h"
#include "slandll.h"

#define STDERRDEBUG

BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
    switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
    }
    return TRUE;
}

typedef struct slanint {
	HANDLE hFile;  // handle to SLANMAC
	HANDLE wMutex; // write mutex, so multiple things can try to write at once.
	OVERLAPPED ol; // for overlapped (blocking) I/O
	struct local_info info;
} SLANINT, *PSLANINT;

SLANDLL_API void *load_slan_interface(DWORD *err)
{
	PSLANINT sl;
	HANDLE h;

	h = CreateFile("\\\\.\\SLANMAC", 
		GENERIC_READ|GENERIC_WRITE,
		0,
		NULL,
		OPEN_EXISTING,
		FILE_FLAG_OVERLAPPED|FILE_FLAG_DELETE_ON_CLOSE,
		0);

	if(h == INVALID_HANDLE_VALUE)
	{
		if(err)
			*err = GetLastError();
		return NULL;
	}

	sl = (PSLANINT) malloc(sizeof(SLANINT));
	if(!sl)
	{
		CloseHandle(h);
		if(err)
			*err = GetLastError();
		return NULL;
	}

	memset(sl, 0, sizeof(SLANINT));
	sl->ol.hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
	sl->wMutex = CreateMutex(NULL, FALSE, NULL);
	sl->hFile = h;

	return (void *) sl;
}

BOOL OpenVxDHandle(HANDLE hEvent, DWORD *vxdhandle)
{
	HINSTANCE hmod;
    FARPROC vfunc = NULL;

    hmod = LoadLibrary("KERNEL32.DLL");
    
    if (hmod)
    {
        vfunc = GetProcAddress(hmod, "OpenVxDHandle");
        if (vfunc)
        {
			DWORD tmp = 0;

            tmp = vfunc(hEvent);
			*vxdhandle = tmp;

			FreeLibrary(hmod);
			return TRUE;
        }
        else
        {
			FreeLibrary(hmod);
			return FALSE;
        }
    }
    return FALSE;
}


SLANDLL_API BOOL start_slan_interface(void *opaque, HANDLE ehandle)
{
	PSLANINT sl = (PSLANINT) opaque;
	DWORD tmp;
	DWORD vxdh;

	if(ehandle && OpenVxDHandle(ehandle, &vxdh))
	{
	   return DeviceIoControl(sl->hFile, IOCTL_SLAN_OPEN, &vxdh, sizeof(DWORD), NULL, 0, &tmp, NULL);	
	}
	else
  	   return DeviceIoControl(sl->hFile, IOCTL_SLAN_OPEN, NULL, 0, NULL, 0, &tmp, NULL);
}

SLANDLL_API BOOL set_slan_info(void *opaque, struct local_info *li)
{
	PSLANINT sl = (PSLANINT) opaque;
	memcpy(&sl->info, li, sizeof(struct local_info) - 16); // do not copy dhcpchaddr!

	return TRUE;
}

SLANDLL_API BOOL stop_slan_interface(void *opaque)
{
	PSLANINT sl = (PSLANINT) opaque;
	DWORD tmp;

	return DeviceIoControl(sl->hFile, IOCTL_SLAN_CLOSE, NULL, 0, NULL, 0, &tmp, NULL);
}

SLANDLL_API void close_slan_interface(void *opaque)
{
	PSLANINT sl = (PSLANINT) opaque;

	CloseHandle(sl->hFile);
	CloseHandle(sl->ol.hEvent);
	CloseHandle(sl->wMutex);
	free(sl);
}

SLANDLL_API BOOL read_slan_interface(void *opaque, char *buffer, DWORD *bufferlength, BOOL wait)
{
	PSLANINT sl = (PSLANINT) opaque;
	OVERLAPPED *overl = NULL;
	BOOL ret;
	DWORD retbytes, err;

	if(wait)
	{
		sl->ol.Offset = 0;
		sl->ol.OffsetHigh = 0;
		ResetEvent(sl->ol.hEvent);
		overl = &sl->ol;
	}

	ret = DeviceIoControl(sl->hFile, IOCTL_SLAN_READ, NULL, 0, buffer, *bufferlength, &retbytes, 
		overl);

	if(!ret)
	{
		err = GetLastError();
		if(err == ERROR_IO_PENDING && overl)
		{
			ret = GetOverlappedResult(sl->hFile, overl, &retbytes, TRUE);
			if(!ret)
				return FALSE;
		}
		else
			return FALSE;
	}

	*bufferlength = retbytes;
	return TRUE;
}

SLANDLL_API BOOL write_slan_interface(void *opaque, char *buffer, DWORD bufferlength)
{
	PSLANINT sl = (PSLANINT) opaque;
	DWORD tmp;
	BOOL ret;

	WaitForSingleObject(sl->wMutex, INFINITE);

	ret = DeviceIoControl(sl->hFile, IOCTL_SLAN_WRITE, buffer, bufferlength, NULL, 0, 
		&tmp, NULL);

	ReleaseMutex(sl->wMutex);

	return ret;
}

/* 
 *
 * DANGER DANGER DANGER WILL ROBINSON
 *
 * I apologize for anyone who has to read the code past this point.
 * The code is rife with magic numbers, stolen from other code that was also rife with
 * magic numbers, or made by myself.
 *
 * Proceed with caution. 
 */


UCHAR remote_ether_addr[6] = {
	0x00, 0x00, 0xCA, 0xFE, 0xF0, 0x0D
};

UCHAR local_ether_addr[6] = {
	0x00, 0x00, 0xDE, 0xAD, 0xBE, 0xEF
};

struct slan_ethernet_header {
	UCHAR toaddr[6];
	UCHAR fromaddr[6];
	UCHAR ptype[2];
};

#pragma pack(push, 1)
struct slan_arp {
	USHORT hardwaretype;
	USHORT protocoltype;
	UCHAR  hardwareaddrlen;
	UCHAR  protocoladdrlen;
	USHORT operation;

	UCHAR  src_hrd_addr[6];
	ULONG  src_ip_addr;
	UCHAR  dst_hrd_addr[6];
	ULONG  dst_ip_addr;
};

struct ipudphdr {
	UCHAR ip_hl:4;
	UCHAR ip_ver:4;
	UCHAR servicetype; 
	USHORT iplen;
	USHORT id;
	USHORT fragoff;
	UCHAR ttl;
	UCHAR protocol;
	USHORT ipchecksum;
	ULONG srcip;
	ULONG dstip;
	USHORT srcport;
	USHORT dstport;
	USHORT udplen;
	USHORT udpchecksum;
};

struct dhcpMessage {
	UCHAR op;
	UCHAR htype;
	UCHAR hlen;
	UCHAR hops;
	UINT xid;
	USHORT secs;
	USHORT flags;
	UINT ciaddr;
	UINT yiaddr;
	UINT siaddr;
	UINT giaddr;
	UCHAR chaddr[16];
	UCHAR sname[64];
	UCHAR file[128];
	UINT cookie;
	UCHAR options[308]; 
};

struct fulldhcpmsg {
	struct slan_ethernet_header eh;
	struct ipudphdr ipu;
	struct dhcpMessage dhcp;
};

#pragma pack(pop)

/* DHCP GARBAGE */
/* op */
#define BOOTREQUEST		1
#define BOOTREPLY		2

/* htype */
#define ETH_10MB		1

/* hlen */
#define ETH_10MB_LEN	6

#define DHCP_MESSAGE_TYPE	0x35

/* Message types */
#define DHCPDISCOVER		1
#define DHCPOFFER			2
#define DHCPREQUEST			3
#define DHCPDECLINE			4
#define DHCPACK				5
#define DHCPNAK				6
#define DHCPRELEASE			7

int DHCPendoption(UCHAR **optionptr) 
{
	UCHAR *tmpptr = *optionptr;
	
	while(tmpptr[0] != 0xff) 
	{
		if(tmpptr[0] == 0x00)
			continue;
		tmpptr++;
		tmpptr += tmpptr[0];
		tmpptr++;
	}
	*optionptr = tmpptr;
	return 0;
}

UCHAR DHCPaddoption(UCHAR *optionptr, UCHAR code,
				char datalength, char *data) 
{
	DHCPendoption(&optionptr);
	optionptr[0] = code;
	if(code == 0xff)
		return 0;
	optionptr++;
	optionptr[0] = datalength;
	optionptr++;
	memcpy(optionptr, data, datalength);
	optionptr += datalength;
	optionptr[0] = 0xff;
	
	return 0;	
}


UCHAR *DHCPgetoption(UCHAR *options, int option_val) 
{
	UCHAR *origoptions = options;

	while(options[0] != 0xff) /* 0xFF is options terminator */
	{
		if(options - origoptions > 306)
			return NULL; /* this is sketchy... look at me later. */

		if(options[0] == 0) /* 0 is options filler. */
		{
			options++;
			continue;
		}

		if(options[0] == 0xff)
			return NULL;

		if(options[0] == option_val)
			return options+2;

		options++;
		options += options[0];
		options++;
	}
	
	return NULL; /* shut up, compiler. bleh! */
}

/* stolen from ISC DHCPD */

UINT checksum (UCHAR *buf, int nbytes, UINT sum) 
{ 
        int i; 

        /* Checksum all the pairs of bytes first... */ 
        for (i = 0; i < (nbytes & ~1); i += 2) { 
                sum += (USHORT) ntohs(*((USHORT *)(buf + i))); 
                /* Add carry. */ 
                if (sum > 0xFFFF) 
                        sum -= 0xFFFF; 
        }         

        /* If there's a single byte left over, checksum it, too.   Network 
           byte order is big-endian, so the remaining byte is the high byte. */ 
        if (i < nbytes) { 
                sum += buf [i] << 8; 
                /* Add carry. */ 
                if (sum > 0xFFFF) 
                        sum -= 0xFFFF; 
        } 
         
        return sum; 
} 

/* Finish computing the sum, and then put it into network byte order. */ 

USHORT wrapsum (UINT sum) 
{ 
     sum = ~sum & 0xFFFF; 

     return htons((USHORT) sum); 
} 

#define DHCPSERVERADDR 0xA9FE6B0A
#define DHCPSERVERADDR_H "\xA9\xFE\x6B\x0A"

int DHCPsendackornak(void *opaque, struct dhcpMessage *oldpacket) 
{
	PSLANINT sl = (PSLANINT) opaque;
	struct dhcpMessage packet;
	int isNAK = 0;
	struct fulldhcpmsg buf;

	memset(&buf, 0, sizeof(struct fulldhcpmsg));
	memset(&packet, 0, sizeof(struct dhcpMessage));
	
	packet.op = BOOTREPLY;
	packet.htype = ETH_10MB;
	packet.hlen = ETH_10MB_LEN;
	packet.xid = oldpacket->xid;
	packet.ciaddr = oldpacket->ciaddr;
	memcpy(&packet.chaddr, oldpacket->chaddr, 16);
	memcpy(&packet.cookie, "\x63\x82\x53\x63", 4);
	memcpy(&packet.options, "\xff", 1);
	
	packet.yiaddr = sl->info.ip;

	if(memcmp(packet.chaddr, sl->info.dhcpchaddr, 16) != 0)
		isNAK = 1;
	
	DHCPaddoption(packet.options, 0x36, 0x04, DHCPSERVERADDR_H);

	if(isNAK)
	{
#ifdef STDERRDEBUG
		fprintf(stderr, "sending NAK for %x\n", packet.yiaddr);
#endif
		DHCPaddoption(packet.options, 0x35, 0x01, "\x06"); /* NAK */
	}
	else
	{
#ifdef STDERRDEBUG
		fprintf(stderr, "sending ACK for %x\n", packet.yiaddr);
#endif
		DHCPaddoption(packet.options, 0x35, 0x01, "\x05"); /* ACK */
		DHCPaddoption(packet.options, 0x33, 0x04, "\x00\x00\x0e\x10"); /* lease time: 3600 secs */
		DHCPaddoption(packet.options, 0x01, 0x04, (char *)&sl->info.netmask); /* subnet */
		DHCPaddoption(packet.options, 0x03, 0x04, (char *)&sl->info.gateway); /* router */
		DHCPaddoption(packet.options, 0x06, 0x04, (char *)&sl->info.ns1); /* dns */
		if(sl->info.ns2)
			DHCPaddoption(packet.options, 0x06, 0x04, (char *)&sl->info.ns2); /* dns */
		if(sl->info.ns3)
			DHCPaddoption(packet.options, 0x06, 0x04, (char *)&sl->info.ns3); /* dns */
		//addOption(packet.options, 0x2C, 0x04, (char *)&inp.s_addr); /* WINS */	
	}

    memcpy(&buf.dhcp, &packet, sizeof(packet));

	buf.ipu.ip_hl = 5;
	buf.ipu.ip_ver = 4;
	buf.ipu.iplen = htons(sizeof(struct fulldhcpmsg) - 14);
	buf.ipu.ttl = 128;
	buf.ipu.protocol = 17;
	buf.ipu.srcip = DHCPSERVERADDR;
	buf.ipu.dstip = packet.yiaddr;
	buf.ipu.srcport = htons(67);
	buf.ipu.dstport = htons(68);
	buf.ipu.udplen = htons(sizeof(struct fulldhcpmsg) - 34);

	buf.ipu.ipchecksum = wrapsum(checksum((UCHAR *)&buf.ipu, 20, 0));
	buf.ipu.udpchecksum = wrapsum(
		checksum((UCHAR *)&buf.ipu.srcport, 8, 
		checksum((UCHAR *)&buf.dhcp, sizeof(struct dhcpMessage), 
		checksum((UCHAR *)&buf.ipu.srcip, 8,
		17 + ntohs(buf.ipu.udplen)))));

#ifdef STDERRDEBUG
	fprintf(stderr, "writing out dhcp ack/nak\n");
#endif

    write_slan_packet_ether(opaque, (char *) &buf, sizeof(struct fulldhcpmsg));

	return 0;
}
 
int DHCPsendoffer(void *opaque, struct dhcpMessage *oldpacket) 
{
	PSLANINT sl = (PSLANINT) opaque;
	struct dhcpMessage packet;
	struct fulldhcpmsg buf;

	memset(&buf, 0, sizeof(struct fulldhcpmsg));
	memset(&packet, 0, sizeof(struct dhcpMessage));
	
	packet.op = BOOTREPLY;
	packet.htype = ETH_10MB;
	packet.hlen = ETH_10MB_LEN;
	packet.xid = oldpacket->xid;

	if((packet.yiaddr = sl->info.ip) == 0) 
	{
		return -1;
	}
	
	memcpy(&packet.chaddr, oldpacket->chaddr, 16);
	memcpy(&sl->info.dhcpchaddr, oldpacket->chaddr, 16);
	memcpy(&packet.cookie, "\x63\x82\x53\x63", 4); /* dhcp magic cookie */
	memcpy(&packet.options, "\xff", 1);
	
	DHCPaddoption(packet.options, 0x35, 0x01, "\x02");	/* dhcpoffer */
	DHCPaddoption(packet.options, 0x36, 0x04, DHCPSERVERADDR_H); /* server IP */
	DHCPaddoption(packet.options, 0x33, 0x04, "\x00\x00\x0e\x10"); /* lease time: 3600 secs */
			
	memcpy(&buf.dhcp, &packet, sizeof(packet));

	buf.ipu.ip_hl = 5;
	buf.ipu.ip_ver = 4;
	buf.ipu.iplen = htons(sizeof(struct fulldhcpmsg) - 14);
	buf.ipu.ttl = 128;
	buf.ipu.protocol = 17;
	buf.ipu.srcip = DHCPSERVERADDR;
	buf.ipu.dstip = packet.yiaddr;
	buf.ipu.srcport = htons(67);
	buf.ipu.dstport = htons(68);
	buf.ipu.udplen = htons(sizeof(struct fulldhcpmsg) - 34);

	buf.ipu.ipchecksum = wrapsum(checksum((UCHAR *)&buf.ipu, 20, 0));
	buf.ipu.udpchecksum = wrapsum(
		checksum((UCHAR *)&buf.ipu.srcport, 8, 
		checksum((UCHAR *)&buf.dhcp, sizeof(struct dhcpMessage), 
		checksum((UCHAR *)&buf.ipu.srcip, 8,
		17 + ntohs(buf.ipu.udplen)))));

#ifdef STDERRDEBUG
	fprintf(stderr, "writing out dhcp offer\n");
#endif

    write_slan_packet_ether(opaque, (char *) &buf, sizeof(struct fulldhcpmsg));

	return 0;
}


static BOOL check_dhcp(void *opaque, char *buf, DWORD buflen)
{
	struct ipudphdr *ipuh;
	char *pkt;
	DWORD pktlen;
	int hlen;
	struct dhcpMessage packet;
	UCHAR *val;

	/* this has an ethernet header, chop it off. */

	buflen -= 14;
	buf += 14;

    if(buflen < sizeof(struct ipudphdr))
		return FALSE;

	ipuh = (struct ipudphdr *) buf;

	if(ipuh->ip_ver != 4) /* not ipv4 */
	{
#ifdef STDERRDEBUG
		fprintf(stderr, "not ipv4 ??? (%d)\n", ipuh->ip_ver);
#endif
		return FALSE;
	}

	if(ipuh->protocol != 17) /* not UDP */
		return FALSE;

	if(ntohs(ipuh->dstport) != 67) /* not to port 67 (bootps) */
		return FALSE;

	if(ipuh->dstip != 0xFFFFFFFF && ipuh->dstip != DHCPSERVERADDR) /* not to 255.255.255.255 */
	{
#ifdef STDERRDEBUG
		fprintf(stderr, "not broadcast or server IP, but bootps??\n");
#endif
		return FALSE;
	}

	hlen = (ipuh->ip_hl << 2) + 8; /* ip header length + udp header length */

	pktlen = buflen - hlen;
	pkt = buf + hlen;

	if(pktlen > sizeof(struct dhcpMessage))
		return FALSE;

	memset(&packet, 0, sizeof(struct dhcpMessage));
	memcpy(&packet, pkt, pktlen);

#ifdef STDERRDEBUG
	{
		struct in_addr ia;
		ia.s_addr = ipuh->srcip;
		fprintf(stderr, "UDP, %s:%d ", inet_ntoa(ia), ntohs(ipuh->srcport));
		ia.s_addr = ipuh->dstip;
		fprintf(stderr, "-> %s:%d\n", inet_ntoa(ia), ntohs(ipuh->dstport));
	}
#endif

	if(ntohl(packet.cookie) != 0x63825363) /* the dhcp magic cookie. isn't that great? */
	{
#ifdef STDERRDEBUG
		fprintf(stderr, "Invalid dhcp magic cookie: %x\n", packet.cookie);
#endif
		return TRUE;
	}

	val = DHCPgetoption(packet.options, DHCP_MESSAGE_TYPE);
	if(!val)
	{
#ifdef STDERRDEBUG
		fprintf(stderr, "Unknown DHCP message type\n");
#endif
		return TRUE;
	}

	switch(val[0])
	{
		case DHCPDISCOVER:
			DHCPsendoffer(opaque, &packet);
			break;

		case DHCPREQUEST:
			DHCPsendackornak(opaque, &packet);
			break;

		default:
#ifdef STDERRDEBUG
			fprintf(stderr, "DHCP message type %d from client (???)\n", val[0]);
#endif
			break;
	}
	return TRUE;
}


#define myhtons(a) ((((a) & 0xFF00) >> 8) | (((a) & 0x00FF) << 8))
#define myntohs(a) (myhtons(a))

#define ARPPKTLEN (sizeof(struct slan_ethernet_header) + sizeof(struct slan_arp))

static int slan_is_inet(void *opaque, char *buf, DWORD buflen)
{
	struct slan_ethernet_header *seh;
	PSLANINT sl = (PSLANINT) opaque;

	if(buflen < 14)
		return FALSE; // what the hell is this? :)

	seh = (struct slan_ethernet_header *) buf;
	if(seh->ptype[0] != 0x08)
		return FALSE; // I have no clue what this is, but it's not NET.

	if(seh->ptype[1] == 0x00)
	{
		/* mm, now we need to check if this is a DHCP packet... */
		if(check_dhcp(opaque, buf, buflen) == TRUE)
			return FALSE; // dhcp handled it... 

		return TRUE; // this is INET!
	}

	if(seh->ptype[1] == 0x06) // this is ARP
	{
		char tbuf[ARPPKTLEN];
		struct slan_ethernet_header *Teth = (struct slan_ethernet_header *) tbuf;
		struct slan_arp *Tarp = (struct slan_arp *) ((char *)tbuf + sizeof(struct slan_ethernet_header));
		struct slan_arp *Sarp = (struct slan_arp *) ((char *)buf + sizeof(struct slan_ethernet_header));

		if(buflen < ARPPKTLEN)
			return FALSE; // what is this? :)

		if(Sarp->hardwareaddrlen != 6)
			return FALSE; // weird ether addr length

		if(Sarp->protocoladdrlen != 4)
			return FALSE; // weird IP length

		if(myntohs(Sarp->hardwaretype) != 1)
			return FALSE; // not ethernet

		if(myntohs(Sarp->operation) != 1)
			return FALSE; // not an ARP request

		if(Sarp->dst_ip_addr == sl->info.ip)
		{
#ifdef STDERRDEBUG
			fprintf(stderr, "Cought arp ping for MY IP!\n");
#endif
			return FALSE; // ARP pings?! AAGH! The pains I go through...
		}

		memcpy(&Teth->fromaddr, &remote_ether_addr, 6);
		memcpy(&Teth->toaddr, &seh->fromaddr, 6);
		Teth->ptype[0] = 0x08;
		Teth->ptype[1] = 0x06;

		Tarp->hardwareaddrlen = 6; // 6 ocets for an ethernet address
		Tarp->hardwaretype = myhtons(1);    // 1 = Ethernet
		Tarp->protocoladdrlen = 4; // 4 ocets for an IP
		Tarp->operation = myhtons(2); // arp reply
		Tarp->protocoltype = myhtons(0x0800); // NET
		memcpy(&Tarp->dst_hrd_addr, &Sarp->src_hrd_addr, 6);
		memcpy(&Tarp->dst_ip_addr, &Sarp->src_ip_addr, 4);
		memcpy(&Tarp->src_ip_addr, &Sarp->dst_ip_addr, 4);
		memcpy(&Tarp->src_hrd_addr, &remote_ether_addr, 6);

		write_slan_interface(opaque, (char *)&tbuf, ARPPKTLEN);

		return FALSE;
	}

	// this is not for the slan client to see...
	return FALSE;
}

struct slan_header {
	DWORD framelength;
	DWORD packetlength;
};

#define myabs(x) ((x) >= 0 ? (x) : -(x))
// returns
// FALSE on done
// TRUE if more data (true means call me again)
// if curpktlen is returned as 0, but returns TRUE, there is still more data
SLANDLL_API BOOL get_slan_packet(void *opaque, char **buffer, DWORD *buflen, char **curbuf, DWORD *curbuflen)
{
	struct slan_header *slh;
	char *curpkt;
	DWORD curpktlen;

	if(*buflen < sizeof(struct slan_header))
	{
		*buflen = 0xFFFFFFFF;
		return FALSE;
	}

	slh = (struct slan_header *) (*buffer);

	if(myabs((int)slh->framelength - (int)slh->packetlength) > 3)
	{
		*buflen = 0xFFFFFFFF;
		return FALSE;
	}

	curpkt = (*buffer) + sizeof(struct slan_header);
	curpktlen = slh->packetlength;

	(*buffer) += (slh->framelength + sizeof(struct slan_header));
	(*buflen) -= (slh->framelength + sizeof(struct slan_header));

	if(slan_is_inet(opaque, curpkt, curpktlen))
	{
		*curbuf = curpkt + 14; // chop off ethernet header
		*curbuflen = curpktlen - 14;
	}
	else
	{
		*curbuf = NULL;
		*curbuflen = 0;
	}

	return *buflen ? TRUE : FALSE;
}


SLANDLL_API BOOL write_slan_packet_ether(void *opaque, char *buf, DWORD bufferlength)
{
	struct slan_ethernet_header *eeth = (struct slan_ethernet_header *) buf;
	memcpy(&eeth->toaddr, &local_ether_addr, 6);
	memcpy(&eeth->fromaddr, &remote_ether_addr, 6);
	eeth->ptype[0] = 0x08;
	eeth->ptype[1] = 0x00;

	return write_slan_interface(opaque, buf, bufferlength);
}
