/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
  $Header: /home/cvsroot/stage4/platform/windows/winclient/winhelp.c,v 1.4 2001/07/06 17:57:38 lucas Exp $
  $Author: lucas $ $Date: 2001/07/06 17:57:38 $ $Name:  $
  $Revision: 1.4 $
 
*/

#include "slan_windows.h"

#include <io.h>
#include <fcntl.h>
#include <commdlg.h>
#include <commctrl.h>
#include <shellapi.h>

#include <stdarg.h>
#include "slan_alloc.h"
#include "slan_types.h"

#include "resource.h"

static HWND h_current_controlling_window;
static HINSTANCE hInst;
static char server_file[512];
static HANDLE wndmutex;
static HANDLE die_event;
static BOOL shutdown_now = FALSE;

#define NUMICONS 6
#define S_ICONWIDTH 32
#define S_ICONHEIGHT 32

typedef struct {
	int selectedicon;
	HICON iconlist[NUMICONS];
	HBRUSH selection_brush;
	HBRUSH nice_blue_background_brush;
	RECT iconrects[NUMICONS];
	RECT textrect;
	HPEN highlightpen;
	char curtext[128];
} slan_start_window_info;

typedef struct {
	char *prompt;
	char *title;
	char *input;
	int inputlen;
	BOOL hide;
} slan_enter_dialog_t;


int gettimeofday(struct timeval *tv, struct timeval *garbage)
{
	SYSTEMTIME st;

	GetSystemTime(&st);

	tv->tv_usec = st.wMilliseconds;
	tv->tv_sec = st.wSecond;

	return 0;
}

int recvall(SOCKET fd, UCHAR *buffer, UINT length, UINT blah)
{
	UINT lenleft = length;
	UINT totallen = 0;

	while(lenleft)
	{
		int ret;

		ret = recv(fd, buffer, lenleft, 0);
		if(ret <= 0)
		{
			printf("recvall failure (%d)\n", ret);
			return ret;
		}

		lenleft -= ret;
		buffer += ret;
		totallen += ret;
	}

	return totallen;
}

void slan_m_log(char *format, ...)
{
	FILE *fp;
	va_list vl;
	char buf[512];

	va_start(vl, format);
	vsprintf(buf, format, vl);
	va_end(vl);

	fp = fopen("c:\\slanmlog.txt", "a");
	if(fp)
	{
#ifdef SLANMLOGSHOWDATE
		time_t now = time(NULL);
		char tmp[512];

		sprintf(tmp, "%s", ctime(&now));
		tmp[strlen(tmp) - 1] = '\0';

		fprintf(fp, "[%s] %s\r\n", tmp, buf);
#else
		fprintf(fp, "%s\r\n", buf);
#endif
		fclose(fp);
	}
}

struct slan_memory {
	char created[128];
	int size;
	int actual_size;
	void *buffer;
	struct slan_memory *next;
};

static struct slan_memory *memhead = NULL;
static int totalmem = 0, maxmem = 0;
static BOOL memory_is_valid;
static HANDLE memory_file, memory_file_mapping;
static void *memory_space;

#define MEMORY_FILE_NAME "slanmem.swp"
#define MEMORY_FILE_SIZE (262144)

void *memp_128[64]; /* 8192 */
void *memp_512[32]; /* 16384 */
void *memp_1024[24]; /* 24576 */
void *memp_10240[10]; /* 102400 */
void *memp_65536[1]; /* 65536 -- special case! (hopefully) */

BOOL mem_init()
{
	unsigned char *memp;
	int i;

	memory_is_valid = FALSE;

	memory_file = CreateFile(MEMORY_FILE_NAME, GENERIC_READ|GENERIC_WRITE,
		0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_DELETE_ON_CLOSE, NULL);

	if(memory_file == NULL)
	{
		fprintf(stderr, "Could not open memory file...\n");
		return FALSE;
	}

	memory_file_mapping = CreateFileMapping(memory_file, NULL, PAGE_READWRITE|SEC_COMMIT, 0, 
		MEMORY_FILE_SIZE, NULL);

	if(memory_file_mapping == NULL)
	{
		fprintf(stderr, "Could not create memory file mapping!\n");
		CloseHandle(memory_file);
		return FALSE;
	}

	memory_space = MapViewOfFile(memory_file_mapping, FILE_MAP_ALL_ACCESS, 0, 0, 0);

	if(memory_space == NULL)
	{
		fprintf(stderr, "Could not map memory file!\n");
		CloseHandle(memory_file_mapping);
		CloseHandle(memory_file);
		return FALSE;
	}

	memp = (unsigned char *) memory_space;

	*(unsigned int *) memp = 0xCACACACA;
	memp += 4;

	for(i = 0; i < 64; i++)
	{
		memp_128[i] = memp;
		memp += 128;
		*(unsigned int *) memp = 0xCACACACA;
		memp += 4;
	}

	for(i = 0; i < 32; i++)
	{
		memp_512[i] = memp;
		memp += 512;
		*(unsigned int *) memp = 0xCACACACA;
		memp += 4;
	}

	for(i = 0; i < 24; i++)
	{
		memp_1024[i] = memp;
		memp += 1024;
		*(unsigned int *) memp = 0xCACACACA;
		memp += 4;
	}

	for(i = 0; i < 10; i++)
	{
		memp_10240[i] = memp;
		memp += 10240;
		*(unsigned int *) memp = 0xCACACACA;
		memp += 4;
	}

	for(i = 0; i < 1; i++)
	{
		memp_65536[i] = memp;
		memp += 65536;
		*(unsigned int *) memp = 0xCACACACA;
		memp += 4;
	}

	memory_is_valid = TRUE;
	return TRUE;
}

void mem_destroy()
{
	if(memory_is_valid)
	{
		memset(memory_space, 0, MEMORY_FILE_SIZE);
		CloseHandle(memory_file_mapping);
		CloseHandle(memory_file);
		memory_is_valid = FALSE;
	}
}

void *get_block(int length, int *actual_size)
{
	int i;
	void *memp;

	if(!memory_is_valid)
		return NULL;

	if(length <= 128)
	{
		for(i = 0; i < 64; i++)
		{
			if(memp_128[i] != NULL)
			{
				*actual_size = 128;
				memp = memp_128[i];
				memp_128[i] = NULL;
				return memp;
			}
		}
	}

	if(length <= 512)
	{
		for(i = 0; i < 32; i++)
		{
			if(memp_512[i] != NULL)
			{
				*actual_size = 512;
				memp = memp_512[i];
				memp_512[i] = NULL;
				return memp;
			}
		}
	}

	if(length <= 1024)
	{
		for(i = 0; i < 24; i++)
		{
			if(memp_1024[i] != NULL)
			{
				*actual_size = 1024;
				memp = memp_1024[i];
				memp_1024[i] = NULL;
				return memp;
			}
		}
	}

	if(length <= 10240)
	{
		for(i = 0; i < 10; i++)
		{
			if(memp_10240[i] != NULL)
			{
				*actual_size = 10240;
				memp = memp_10240[i];
				memp_10240[i] = NULL;
				return memp;
			}
		}
	}

	if(length == 65536) /* special case: there is only one place in the code where this happens */
	{
		for(i = 0; i < 1; i++)
		{
			if(memp_65536[i] != NULL)
			{
				*actual_size = 65536;
				memp = memp_65536[i];
				memp_65536[i] = NULL;
				return memp;
			}
		}
	}

	fprintf(stderr, "Out of memory or alloc too large (%d)! resorting to malloc()!\n",
		length);

	*actual_size = -1;
	memp = malloc(length);

	if(!memp)
	{
		*actual_size = 0;
		fprintf(stderr, "really out of memory!\n");
		return NULL;
	}

	return memp;
}

BOOL free_block(void *mem, int actual_size)
{
	int i, num;
	unsigned char *sigcheck;
	void **bptr;

	if(actual_size == -1)
	{
		free(mem);
		return TRUE;
	}

	memset(mem, 0, actual_size);

	switch(actual_size)
	{
	case 128:
		num = 64;
		bptr = memp_128;
		break;

	case 512:
		num = 32;
		bptr = memp_512;
		break;

	case 1024:
		num = 24;
		bptr = memp_1024;
		break;

	case 10240:
		num = 10;
		bptr = memp_10240;
		break;

	case 65536:
		num = 1;
		bptr = memp_65536;
		break;

	default:
		fprintf(stderr, "free_block for invalid size %d.. memory corruption??\n", actual_size);
		return FALSE;
	}

	sigcheck = (unsigned char *) mem;
	sigcheck -= 4;
	if(*(unsigned int *)sigcheck != 0xCACACACA)
		fprintf(stderr, "Memory before block %x (%d) is smashed!\n", (unsigned char *) mem, actual_size);

	sigcheck = (unsigned char *) mem;
	sigcheck += actual_size;
	if(*(unsigned int *)sigcheck != 0xCACACACA)
		fprintf(stderr, "Memory after block %x (%d) is smashed!\n", (unsigned char *) mem, actual_size);

	for(i = 0; i < num; i++)
	{
		if(bptr[i] == NULL)
		{
			bptr[i] = mem;
			return TRUE;
		}
	}

	fprintf(stderr, "umm.. I have a homeless block of %d bytes. This is not good!\n", actual_size);
	return FALSE;
}

static void *add_memory(int size
#ifdef MEMDEBUG
					   , 
					   char *location
#endif
					   )
{
	struct slan_memory *slm = (struct slan_memory *) malloc(sizeof(struct slan_memory));
	slm->size = size;
#ifdef MEMDEBUG
	strncpy(slm->created, location, 128);
#else
	strncpy(slm->created, "debug off!", 128);
#endif
	slm->buffer = get_block(size, &slm->actual_size);
	slm->next = memhead;
	memhead = slm;

	totalmem += size;
	if(totalmem > maxmem)
		maxmem = totalmem;
#ifdef MEMDEBUG_LOUD
	fprintf(stderr, "Total allocated memory (+%d): %d max %d\n", size, totalmem, maxmem);
#endif

	return slm->buffer;
}

static void del_memory(
					   void *orig
#ifdef MEMDEBUG
					   ,
					   char *location
#endif
					   )
{
	struct slan_memory *slm, *slmp = NULL;

	for(slm = memhead; slm; slm = slm->next)
	{
		if(slm->buffer == orig)
			break;
		slmp = slm;
	}

	if(!slm)
	{
		fprintf(stderr, "del_memory(%x) -- block nonexistant!\n", orig);
#ifdef MEMDEBUG
		fprintf(stderr, "nonexistant free from %s\n", location);
		slan_m_log("del_memory(%x) freeing nonexistant block at %s", orig, location); 
#endif
		return;
	}

	if(slmp)
		slmp->next = slm->next;
	else
		memhead = slm->next;

	totalmem -= slm->size;
	free_block(slm->buffer, slm->actual_size);
#ifdef MEMDEBUG_LOUD
	fprintf(stderr, "Total allocated memory (-%d): %d\n", slm->size, totalmem);
#endif
	free(slm);
}

#ifdef MEMDEBUG
void *slan_mallocd(int size, char *file, int line)
#else
void *slan_malloc(int size)
#endif
{
	void *ret;
#ifdef MEMDEBUG
	char locationbuf[128];
	_snprintf(locationbuf, 128, "%s (%d)", file, line);
	ret = add_memory(size, locationbuf);
#else
	ret = add_memory(size);
#endif
	return ret;
}

#ifdef MEMDEBUG
void *slan_callocd(int num, int size, char *file, int line)
#else
void *slan_calloc(int num, int size)
#endif
{
	void *ret;
#ifdef MEMDEBUG
	char locationbuf[128];
	_snprintf(locationbuf, 128, "%s (%d)", file, line);
	ret = add_memory(size * num, locationbuf);
#else
	ret = add_memory(size * num);
#endif
	if(ret) memset(ret, 0, size * num);
	return ret;
}

#ifdef MEMDEBUG
void *slan_reallocd(void *orig, int size, char *file, int line)
#else
void *slan_realloc(void *orig, int size)
#endif
{
	void *ret;
#ifdef MEMDEBUG
	char locationbuf[128];
	_snprintf(locationbuf, 128, "%s (%d)", file, line);
	ret = add_memory(size, locationbuf);
#else
	ret = add_memory(size);
#endif
	if(ret)
	{
		memcpy(ret, orig, size);
		slan_memfree(orig);
	}
	return ret;
}

#ifdef MEMDEBUG
void slan_memfreed(void *orig, char *file, int line)
#else
void slan_memfree(void *orig)
#endif
{
#ifdef MEMDEBUG
	char locationbuf[128];
	_snprintf(locationbuf, 128, "%s (%d)", file, line);
	del_memory(orig, locationbuf);
#else
	del_memory(orig);
#endif
}

void memory_cleanup()
{
	struct slan_memory *slm;

	while((slm = memhead))
	{
		slan_m_log("Memory leak: %d bytes: allocated at %s", 
			slm->size, slm->created);
		fprintf(stderr, "Memory leak: %d bytes : allocated at %s\n",
			slm->size, slm->created);
		del_memory(slm->buffer, "memory_cleanup (0)");
	}
}

//////////////////////////////
//////////////////////////////

INT_PTR CALLBACK SlanEnterProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	static slan_enter_dialog_t *sedt = NULL;

	switch(msg)
	{
		case WM_INITDIALOG:
			{
				sedt = (slan_enter_dialog_t *) lparam;

				SetWindowText(hwnd, sedt->title);
				SetDlgItemText(hwnd, IDC_MESSAGE, sedt->prompt);
				SendMessage(GetDlgItem(hwnd, IDC_VALUE), EM_LIMITTEXT, sedt->inputlen, 0);
				if(sedt->hide)
					SendMessage(GetDlgItem(hwnd, IDC_VALUE), EM_SETPASSWORDCHAR, '#', 0);
				else
					SendMessage(GetDlgItem(hwnd, IDC_VALUE), EM_SETPASSWORDCHAR, 0, 0);
				SetFocus(GetDlgItem(hwnd, IDC_VALUE));
			}
			break;

		case WM_COMMAND:
			switch(LOWORD(wparam))
			{
				case IDOK:
					if(sedt)
					{
						GetDlgItemText(hwnd, IDC_VALUE, sedt->input, sedt->inputlen);
						sedt = NULL;
					}
					EndDialog(hwnd, 0);

				default:
					break;
			}

		default:
			break;
	}

	return FALSE;
}

unsigned int get_data(HWND parentwnd, char *prompt, char *title, char *input, int inputlen, BOOL hide)
{
   slan_enter_dialog_t sedt;

   sedt.prompt = prompt;
   sedt.title = title;
   sedt.input = input;
   sedt.inputlen = inputlen;
   sedt.hide = hide;

   return DialogBoxParam(hInst, MAKEINTRESOURCE(IDD_SLANENTER), parentwnd, SlanEnterProc, (LPARAM) &sedt);
}

static int get_new_configfile(HWND parentwnd)
{
	OPENFILENAME ofn;

	memset(&ofn, 0, sizeof(OPENFILENAME));

	ofn.hInstance = hInst;
	ofn.hwndOwner = parentwnd;
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.lpstrFilter = "SLAN Configuration Files\0*.SLC\0Text Files\0*.TXT\0\0";
	ofn.nFilterIndex = 1;
	ofn.lpstrFile = server_file;
	ofn.nMaxFile = 512;
	ofn.lpstrTitle = "Select SLAN Configuration File";
	ofn.Flags = OFN_FILEMUSTEXIST |
		OFN_HIDEREADONLY |
		OFN_NONETWORKBUTTON |
		OFN_PATHMUSTEXIST;

	return GetOpenFileName(&ofn);
}

static char *filename_only(char *fn)
{
	char *x = fn + strlen(fn) - 1;

	while(x > fn && *x != '\\')
		x--;

	if(*x != '\\')
		return "no file selected";

	return x + 1;
}

INT_PTR CALLBACK SlanConnectProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	switch(msg)
	{
		case WM_INITDIALOG:
			{
				int retval;
				char tmp[2];

				memset(server_file, 0, 512);

				tmp[0] = '\0';

				retval = GetPrivateProfileString("SlanConnect", "ServerFile", tmp, server_file,
					512, "slan.ini");

				if(retval != 0)
				{
					SetDlgItemText(hwnd, IDC_CONFIGFILE, filename_only(server_file));
				}

				SendDlgItemMessage(hwnd, IDC_DEBUG, BM_SETCHECK, BST_CHECKED, 0);

			}
			break;

		case WM_CLOSE:
			SendMessage(hwnd, WM_COMMAND, IDCANCEL, 0);
			break;

		case WM_COMMAND:
			switch(LOWORD(wparam))
			{
				case IDC_CHANGESERVER:
					if(get_new_configfile(hwnd))
						SetDlgItemText(hwnd, IDC_CONFIGFILE, filename_only(server_file));
					break;

				case IDOK:
					{
						UINT checkval;

						if(*server_file == '\0')
						{
							MessageBox(hwnd, "No configuration file selected", "Error",
								MB_OK|MB_ICONSTOP);
							break;
						}

						WritePrivateProfileString("SlanConnect", "ServerFile", server_file, "slan.ini");

						checkval = SendDlgItemMessage(hwnd, IDC_DEBUG, BM_GETCHECK, 0, 0);

						EndDialog(hwnd, checkval == BST_CHECKED ? 2 : 1);
					}
					break;

				case IDCANCEL:
					EndDialog(hwnd, 0);
					break;

				default:
					break;
			}
			break;
	}

	return FALSE;
}

////////////////////////////////////////////////
HWND get_current_main_window()
{
	WaitForSingleObject(wndmutex, INFINITE);
	return h_current_controlling_window;
}

void release_current_main_window()
{
	ReleaseMutex(wndmutex);
}

void new_slan_status(int stage, char *text, int destroy)
{
	HWND h = get_current_main_window();

	if(destroy && h)
	{
		PostMessage(h, WM_USER+2, 0, 0);
		h_current_controlling_window = NULL;
	}
	else if(h) 
		SendMessage(h, WM_USER+1, (WPARAM) stage, (LPARAM) text);

	release_current_main_window();
}

void stop_all_gui()
{
	HWND h = get_current_main_window();
	if(h)
	{
		PostMessage(h, WM_USER+2, 0, 0);
		h_current_controlling_window = NULL;
	}
	shutdown_now = TRUE;
	release_current_main_window();
}

LONG APIENTRY SLANStartWindowProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static slan_start_window_info *swi;
	switch(msg)
	{
		case WM_CREATE:
			{
				int i;

				swi = calloc(1, sizeof(slan_start_window_info));
				strcpy(swi->curtext, "...");

				swi->selectedicon = 0;
				for(i = 0; i < NUMICONS; i++)
					swi->iconlist[i] = LoadImage(hInst, MAKEINTRESOURCE(IDI_SLAN), IMAGE_ICON, 
					0, 0, LR_DEFAULTCOLOR);

				for(i = 0; i < NUMICONS; i++)
				{
					int x;
					int y = 20;
					RECT r;

					x = 20 + (i * (S_ICONWIDTH * 2));

					r.top = y - (S_ICONHEIGHT / 4);
					r.bottom = y + S_ICONHEIGHT + (S_ICONHEIGHT / 4);
					r.right = x + S_ICONWIDTH + (S_ICONWIDTH / 4);
					r.left = x - (S_ICONWIDTH / 4);

					swi->iconrects[i] = r;
				}

				GetClientRect(hwnd, &swi->textrect);
				swi->textrect.top = 75;
				swi->textrect.bottom = 100;
				swi->textrect.left += 20;
				swi->textrect.right -= 20;

				swi->selection_brush = CreateSolidBrush(RGB(0xFF, 0xFF, 0xFF));
				swi->nice_blue_background_brush = CreateSolidBrush(RGB(0x44, 0x44, 0x66));

				swi->highlightpen = CreatePen(PS_INSIDEFRAME, 2, RGB(0, 0, 0xFF));

			}
			break;

		case WM_CLOSE:
			DestroyWindow(hwnd);
			break;

			/*
		case WM_TIMER:
			{
				RECT r;

				InvalidateRect(hwnd, &swi->iconrects[swi->selectedicon], TRUE);

				if(++swi->selectedicon == NUMICONS)
					swi->selectedicon = 0;

				InvalidateRect(hwnd, &swi->iconrects[swi->selectedicon], TRUE);
				InvalidateRect(hwnd, &swi->textrect, TRUE);
			}
			break;
			*/

		case WM_USER+1:
			{
				/* note that the mutex is already held here! */

				if(!swi)
					break;

				if(wParam != 0xFFFFFFFF && wParam < NUMICONS)
				{
					InvalidateRect(hwnd, &swi->iconrects[swi->selectedicon], TRUE);
					swi->selectedicon = wParam;
				}
				strncpy(swi->curtext, (char *) lParam, 128);

				InvalidateRect(hwnd, &swi->iconrects[swi->selectedicon], TRUE);
				InvalidateRect(hwnd, &swi->textrect, TRUE);
			}
			break;

		case WM_USER+2:
			DestroyWindow(hwnd);
			break;

		case WM_DESTROY:
			{
				int i;

				for(i = 0; i < NUMICONS; i++)
					DestroyIcon(swi->iconlist[i]);
				DeleteObject(swi->selection_brush);
				DeleteObject(swi->highlightpen);
				DeleteObject(swi->nice_blue_background_brush);
				free(swi);
				swi = NULL;
			}
			PostQuitMessage(0);
			break;

		case WM_PAINT:
			{
				HDC hdc;
				PAINTSTRUCT ps;
				int i;
				int bkmode;

				COLORREF origcolor;

				hdc = BeginPaint(hwnd, &ps);

				for(i = 0; i < NUMICONS; i++)
				{
					int x;
					int y = 20;

					x = 20 + (i * (S_ICONWIDTH * 2));

					if(swi->selectedicon == i)
					{
						HGDIOBJ oldobj = SelectObject(hdc, swi->highlightpen);
						RECT r = swi->iconrects[i];

						MoveToEx(hdc, r.left+1, r.top+1, NULL);
						LineTo(hdc, r.right-1, r.top+1);
						LineTo(hdc, r.right-1, r.bottom-1);
						LineTo(hdc, r.left+1, r.bottom-1);
						LineTo(hdc, r.left+1, r.top+1);

						SelectObject(hdc, oldobj);
					//	FillRect(hdc, &iconrects[i], selection_brush);
					}

					DrawIcon(hdc, x, y, swi->iconlist[i]);
				}

				bkmode = SetBkMode(hdc, TRANSPARENT);

				origcolor = SetTextColor(hdc, RGB(0xFF, 0xFF, 0xFF));

				DrawTextEx(hdc, swi->curtext, -1, 
					&swi->textrect,
					DT_CENTER|DT_TOP, NULL);

				SetTextColor(hdc, origcolor);
				SetBkMode(hdc, bkmode);

				EndPaint(hwnd, &ps);
				return 0;
			}

		case WM_ERASEBKGND:
		{
			HDC hdc = (HDC) wParam;
			RECT r;
			HGDIOBJ oldobj = SelectObject(hdc, swi->nice_blue_background_brush);

			GetClipBox(hdc, &r);
			PatBlt(hdc, r.left, r.top, r.right-r.left, r.bottom-r.top, PATCOPY);

			SelectObject(hdc, oldobj);

			return TRUE;
		}

		default:
			return DefWindowProc(hwnd, msg, wParam, lParam);
	}
	return FALSE;
}

LONG APIENTRY SLANIdleWindowProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static NOTIFYICONDATA slan_nid;
	static HMENU imenu;

	switch(msg)
	{
		case WM_CREATE:
			memset(&slan_nid, 0, sizeof(NOTIFYICONDATA));

			imenu = LoadMenu(hInst, MAKEINTRESOURCE(IDR_ICONMENU));

			slan_nid.hIcon = LoadIcon(hInst, MAKEINTRESOURCE(IDI_SLAN));
			slan_nid.cbSize = sizeof(NOTIFYICONDATA);
			slan_nid.hWnd = hwnd;
			slan_nid.uCallbackMessage = WM_USER + 3;
			slan_nid.uID = 100;
			slan_nid.uFlags = NIF_ICON|NIF_TIP|NIF_MESSAGE;
			strcpy(slan_nid.szTip, "SLAN up and running...");

			Shell_NotifyIcon(NIM_ADD, &slan_nid);
			break;

		case WM_USER+2:
			DestroyWindow(hwnd);
			break;

		case WM_USER+3:
			if(lParam == WM_RBUTTONDOWN)
			{
				POINT pt;

				/* this looks like an error, but it's a workaround for a win32 bug */
				SetForegroundWindow(hwnd);

				GetCursorPos(&pt);
				TrackPopupMenu(GetSubMenu(imenu, 0), TPM_RIGHTBUTTON, pt.x, pt.y, 0, hwnd, NULL);

				/* workaround part #2 */
				PostMessage(hwnd, WM_NULL, 0, 0);
			}
			break;

		case WM_DESTROY:
			Shell_NotifyIcon(NIM_DELETE, &slan_nid);
			PostQuitMessage(0);
			break;

		case WM_CLOSE:
			DestroyWindow(hwnd);
			break;

		case WM_COMMAND:
			switch(LOWORD(wParam))
			{
				case MMAIN_EXIT:
					SetEvent(die_event);
					PostMessage(hwnd, WM_CLOSE, 0, 0);
					break;

				default:
					break;
			}
			break;

		default:
			return DefWindowProc(hwnd, msg, wParam, lParam);
	}

	return FALSE;
}

////////////////////////////////////////////////

BOOL APIENTRY InitClasses(HINSTANCE hinstance)
{
    WNDCLASS	wc;

	memset(&wc, 0, sizeof(WNDCLASS));

    /* Register the main window class */
    wc.style	     = 0;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance	 = hinstance;
    wc.hCursor	     = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = NULL; 
    wc.lpfnWndProc   = (WNDPROC) SLANStartWindowProc;
    wc.hIcon	     = NULL; 
    wc.lpszMenuName  = NULL;
    wc.cbWndExtra    = 0;
    wc.lpszClassName = "SlanStartWindowClass";
    if (!RegisterClass(&wc))
		return FALSE;

	memset(&wc, 0, sizeof(WNDCLASS));

    wc.style	     = 0;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance	 = hinstance;
    wc.hCursor	     = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = NULL; 
    wc.lpfnWndProc   = (WNDPROC) SLANIdleWindowProc;
    wc.hIcon	     = NULL; 
    wc.lpszMenuName  = NULL;
    wc.cbWndExtra    = 0;
    wc.lpszClassName = "SlanIdleWindowClass";
    if (!RegisterClass(&wc))
		return FALSE;

	return TRUE;
}

void do_initialization(HINSTANCE hInstance)
{
	INITCOMMONCONTROLSEX icex;
	WSADATA wsd;

	icex.dwICC = ICC_WIN95_CLASSES;
	icex.dwSize = sizeof(INITCOMMONCONTROLSEX);

	InitCommonControlsEx(&icex);
	InitClasses(hInstance);

	WSAStartup(0x0202, &wsd);
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
	UINT retval;
	HWND hwnd;
	MSG msg;
	HANDLE slthread;

	hInst = hInstance;

	do_initialization(hInstance);
	h_current_controlling_window = NULL;

	retval = DialogBox(hInstance, MAKEINTRESOURCE(IDD_SLANCONNECT), NULL, SlanConnectProc);

	if(retval == 0)
		return 0;

#ifndef NO_DEBUG_CONSOLE
	if(retval == 2) /* debug on! */
	{
	/* This undocumented, obscure win32 voodoo magic sets up a console, remapping stderr
	and stdout... */

	int hCrt;
	FILE *hf;

	AllocConsole();

	hCrt = _open_osfhandle((long) GetStdHandle(STD_OUTPUT_HANDLE), _O_TEXT);
	hf = _fdopen( hCrt, "w" );
	*stdout = *hf;
	setvbuf(stdout, NULL, _IONBF, 0); 

	hCrt = _open_osfhandle((long) GetStdHandle(STD_ERROR_HANDLE), _O_TEXT);
	hf = _fdopen(hCrt, "w");
	*stderr = *hf;
	setvbuf(stderr, NULL, _IONBF, 0); 
	}
#endif

	wndmutex = CreateMutex(NULL, FALSE, NULL);
	die_event = CreateEvent(NULL, FALSE, FALSE, NULL);

	hwnd = CreateWindowEx(0, "SlanStartWindowClass", "SLAN Client",
		WS_POPUP,
		100,	/* x */
		100,				/* y */
		420,	/* width */
		120,	/* height */
		NULL,			/* parent */
		NULL,			/* menu */
		hInstance,		/* instance */
		NULL);			/* creation data */

	WaitForSingleObject(wndmutex, INFINITE);
	h_current_controlling_window = hwnd;
	ReleaseMutex(wndmutex);

	slthread = start_slan_thread(server_file, die_event);

	ShowWindow(hwnd, SW_SHOW);

    while(GetMessage(&msg, NULL, 0, 0))
	{
	    TranslateMessage(&msg);
	    DispatchMessage(&msg);
	}

	if(!shutdown_now)
	{

		WaitForSingleObject(wndmutex, INFINITE);

		hwnd = CreateWindowEx(0, "SlanIdleWindowClass", "SLAN Client",
			WS_POPUP,
			0,	/* x */
			0,				/* y */
			0,	/* width */
			0,	/* height */
			NULL,			/* parent */
			NULL,			/* menu */
			hInstance,		/* instance */
			NULL);			/* creation data */

		h_current_controlling_window = hwnd;
		ReleaseMutex(wndmutex);

		while(GetMessage(&msg, NULL, 0, 0))
		{
		    TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	SetEvent(die_event);
	WaitForSingleObject(slthread, 15000);
	GetExitCodeThread(slthread, &retval);
	if(retval == STILL_ACTIVE)
	{
		fprintf(stderr, "Closing thread.. aaaargh!\n");
		TerminateThread(slthread, 0);
	}
	CloseHandle(slthread);

	CloseHandle(wndmutex);
	CloseHandle(die_event);

#ifndef NO_DEBUG_CONSOLE
	MessageBox(NULL, "SLAN client is exiting", "Preserving Debug Console", MB_OK);
#endif

	WSACleanup();
	return 0;
}
