/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
  $Header: /home/cvsroot/stage4/platform/windows/winclient/winslan_link.c,v 1.2 2001/07/06 17:08:26 lucas Exp $
  $Author: lucas $ $Date: 2001/07/06 17:08:26 $ $Name:  $
  $Revision: 1.2 $
 
*/

#include "slan_windows.h"
#include "slan_alloc.h"
#include "slan_types.h"

#include "slandll.h"

#include "proto_stack.h"

#include "slan_version.h"

typedef struct {
  void *opaquedriver;
  char *buffer;
} slan_link_t;


/* This gets a message from the slan device. If a message is
   available, it sends it through the protocol stack. If a message is
   not available, it simply exits the protocol stack. */
static void slan_link_down(proto_stack_t *stack, slan_message_t
			   *message) 
{
  slan_link_t *local;
  char *buffer;
  int bufferlength = 65536;
  unsigned char *data_orig_ptr = message->data;

  local = (slan_link_t *) stack->data;
  buffer = local->buffer;

  if(!read_slan_interface(local->opaquedriver, buffer, &bufferlength, FALSE))
  {
	  fprintf(stderr, "Error reading slan packet!\n");
	  slan_free(message);
	  return;
  }
  else if(bufferlength == 0)
  {
	  slan_free(message);
	  return;
  }
  else
  {
	  char *mainbuf = buffer;
	  DWORD mainbuflen = bufferlength;
	  char *curbuf;
	  DWORD curlen;
	  BOOL cont;

	  while(1)
	  {
		  cont = get_slan_packet(local->opaquedriver, &mainbuf, &mainbuflen, &curbuf, &curlen);
		  //printf("packet, %d bytes\n", curlen);

		  // mainbuflen is -1 if there was an internal error. This is a bad thing (tm)
		  // and should never really happen.

		  if(mainbuflen == 0xFFFFFFFF)
			  break;

		  // the current 'packet' could have been processed by the DLL layer.
		  // If it was, ignore it. 

		  if(curlen == 0)
		  {
			  if(cont)
				  continue;
		  }
		  else
		  {
			  if((int) curlen > message->maxspace)
			  {
				 fprintf(stderr, "dropping packet, curlen > message->maxspace\n");
			  }
			  else
			  {
				  message->header = message->data = data_orig_ptr;
				  message->trailer = message->data;
				  memcpy(message->data, curbuf, curlen);
				  message->length = curlen;
				  stack->next->down(stack->next, message);
			  }
		  }

		  if(!cont)
			  break;	
	  }
	
	  if(mainbuflen == 0xFFFFFFFF)
	  {
		  fprintf(stderr, "Strange internal slandll error...\n");
	  }
  }
}

static void slan_link_up(proto_stack_t *stack, slan_message_t
			 *message) {
  slan_link_t *local;

  local = (slan_link_t *) stack->data;

  // reserve the 14 bytes for the API to use...
  slan_reserve(message, 14);
  slan_set(message);

  if(!write_slan_packet_ether(local->opaquedriver, message->data, message->length))
  {
	  fprintf(stderr, "error writing to slan interface\n");
	  return;
  }

  stack->prev->up(stack->prev, message);
}

int winslan_link_register(proto_stack_t *stack_entry, void *slan_opaque) 
{
  slan_link_t *local;

  stack_entry->up = slan_link_up;
  stack_entry->down = slan_link_down;
  stack_entry->header_reserve = 14; // fake ethernet header length
  stack_entry->trailer_reserve = 0;

  local = (slan_link_t *) slan_calloc(1, sizeof(slan_link_t));
  if (local == NULL) return -1;

  local->opaquedriver = slan_opaque;
  local->buffer = slan_malloc(65536);
  if(local->buffer == NULL)
	  return -1;

  stack_entry->data = local;

  return 0;
}
